"""
Benchmarks for 'merge' functions.
"""
__all__ = [
    "bench_merge",
    "bench_merge2",
    "bench_merge2_with_settings",
    "bench_merge_pandas",
    "compare_merge",
]

import itertools
import logging
from itertools import product
from typing import List, Tuple

import numpy as np
from numpy.random import default_rng

from ..rt_categorical import Categorical
from ..rt_dataset import Dataset
from ..rt_merge import merge, merge2
from ..rt_numpy import empty
from .benchmark import timestamper
from .rand_data import rand_dataset
from .runner import benchmark, create_comparison_dataset, create_trial_dataset

logger = logging.getLogger(__name__)
"""The logger for this module."""


def _range_slice(s: slice, endpoint: bool = True) -> List[int]:
    """
    Create a list of integers as specified by the slice.

    Parameters
    ----------
    s : slice
    endpoint : bool

    Returns
    -------
    slice_indices : list of int

    Examples
    --------
    >>> _range_slice(slice(200, 500, 50))
    [200,
     250,
     300,
     350,
     400,
     450,
     500]
    """
    # If the caller wants the right endpoint to be included, adjust
    # the 'stop' value by 1 (or -1) so the .indices() method includes it.
    if endpoint:
        adjust = 1 if s.step >= 0 else -1
        s = slice(s.start, s.stop + adjust, s.step)
    return list(range(*(s.indices(s.stop))))


def generate_merge_datasets(
    rng_seed=12345,
    left_key_unique_count=25,
    right_key_unique_count=2500,
    left_dataset_max_rowcount=250_000,
    right_dataset_max_rowcount=62_500,
) -> Tuple[Dataset, Dataset]:
    """Generates the left and right Datasets for merge benchmarking."""
    left_step = int(left_dataset_max_rowcount / 4)
    left_rowcounts = list(range(left_step, left_dataset_max_rowcount + 1, left_step))
    right_step = int(right_dataset_max_rowcount / 4)
    right_rowcounts = list(range(right_step, right_dataset_max_rowcount + 1, right_step))

    left_datasets, right_datasets = list(), list()
    for left_rowcount in left_rowcounts:
        for right_rowcount in right_rowcounts:
            rng = default_rng(rng_seed)
            left_datasets.append(rand_dataset(left_rowcount, rng, left_key_unique_count))
            right_datasets.append(rand_dataset(right_rowcount, rng, right_key_unique_count))
    return left_datasets, right_datasets


def bench_merge(**kwargs) -> Dataset:
    # TODO: Add additional dimensions:
    #  * key is Cat [True, False]
    #  * key dtype(s) -- use this to also bench with multiple keys
    #  * number of threads
    #  * recycler on/off
    #  * number of additional (non-key) columns in the left and/or right (just make them all the same dtype, it doesn't matter much for this benchmark)
    #  * (advanced) different key distributions
    #  * (advanced) key clustering (i.e. are keys already more or less occurring near each other, or are they all spread out?)
    #  * (advanced) key dispersion -- if the unique keys are created like arange(1, 100), does grouping/merge go any
    #       faster than if the keys are created like arange(1, 1000, 10) or arange(1, 1_000_000_000, 10_000_000_000)?
    #       Make sure to control for dtype to ensure that's the same for all cases tested here.
    #       This could be used to diagnose issues with hashing/grouping implementations.

    # Fixed parameters which apply to all of the trials in this benchmark.
    warmup_iters = 1
    iters = 5

    # Setup parameters.
    rng_seeds = [12345]
    left_key_unique_counts = [100]
    right_key_unique_counts = [10000]
    left_rowcounts = _range_slice(slice(500_000, 2_000_000, 500_000))
    right_rowcounts = _range_slice(slice(250_000, 500_000, 250_000))

    setup_params = itertools.product(
        rng_seeds,
        left_key_unique_counts,
        right_key_unique_counts,
        left_rowcounts,
        right_rowcounts,
    )

    # Trial parameters.
    hows = ["left", "right", "inner"]  # TODO: Add 'outer'; seems to be broken though?
    trial_params = hows

    # Datasets containing timing data and parameters from the trials in this benchmark.
    benchmark_data: List[Dataset] = []

    for (
        rng_seed,
        left_key_unique_count,
        right_key_unique_count,
        left_rowcount,
        right_rowcount,
    ) in setup_params:
        #
        # Setup phase. The data here is used for both the warmup and the real, timed function invocations.
        #
        # Make sure to re-initialize the RNG each time so we get a repeatable result.
        rng = default_rng(rng_seed)

        left_ds = rand_dataset(left_rowcount, rng, left_key_unique_count)
        right_ds = rand_dataset(right_rowcount, rng, right_key_unique_count)

        # Sweep over trial parameters; all trials sharing the same setup parameters
        # share the same data created from those setup parameters.
        for how in trial_params:
            # Allocate an array to hold the raw timing data.
            # TODO: Change to use TimeSpan?
            timing_data = empty(iters, dtype=np.int64)

            for is_warmup in (True, False):
                loop_count = warmup_iters if is_warmup else iters
                logger.info(
                    f"bench_merge:\tmode={'warmup' if is_warmup else 'bench'}\tloops={loop_count}\tleft_rowcount={left_rowcount}\tright_rowcount={right_rowcount}\thow={how}"
                )

                for i in range(loop_count):
                    start_time_ns = timestamper()

                    ### The actual function invocation ###
                    # TODO: This could be passed in as a lambda, and we call it with the data constructed
                    #  in the setup phase + any "parameters" that weren't used in the setup phase; or for
                    #  simplicity just pass all the parameters + the data constructed in the setup phase,
                    #  the lambda doesn't need to use all of it.

                    merge(left_ds, right_ds, on="key", how=how)

                    ### Store the timing results (if this was a real invocation).
                    call_nanos = timestamper() - start_time_ns
                    if not is_warmup:
                        timing_data[i] = call_nanos

            # Create a mini Dataset with the timing results for this run.
            # Capture the timing results along with the other options used for the function invocations.
            trial_data = create_trial_dataset(
                timing_data,
                {
                    # Setup parameters
                    "rng_seed": rng_seed,
                    "left_key_unique_count": left_key_unique_count,
                    "right_key_unique_count": right_key_unique_count,
                    "left_rowcount": left_rowcount,
                    "right_rowcount": right_rowcount,
                    # Trial parameters
                    "how": how,
                },
            )
            benchmark_data.append(trial_data)

    # hstack all of the individual Datasets together into one large Dataset and return it.
    return Dataset.hstack(benchmark_data, destroy=True)


def bench_merge2(**kwargs) -> Dataset:
    # TODO: Add additional dimensions:
    #  * key is Cat [True, False]
    #  * key dtype(s) -- use this to also bench with multiple keys
    #  * number of threads
    #  * recycler on/off
    #  * number of additional (non-key) columns in the left and/or right (just make them all the same dtype, it doesn't matter much for this benchmark)
    #  * (advanced) different key distributions
    #  * (advanced) key clustering (i.e. are keys already more or less occurring near each other, or are they all spread out?)
    #  * (advanced) key dispersion -- if the unique keys are created like arange(1, 100), does grouping/merge go any
    #       faster than if the keys are created like arange(1, 1000, 10) or arange(1, 1_000_000_000, 10_000_000_000)?
    #       Make sure to control for dtype to ensure that's the same for all cases tested here.
    #       This could be used to diagnose issues with hashing/grouping implementations.

    # Fixed parameters which apply to all of the trials in this benchmark.
    warmup_iters = 1
    iters = 5

    # Setup parameters.
    rng_seeds = [12345]
    left_key_unique_counts = [100]
    right_key_unique_counts = [10000]
    left_rowcounts = _range_slice(
        slice(500_000, 2_000_000, 500_000)
    )  # TODO: For larger-memory machines, we could use a larger max here, e.g. 7M
    right_rowcounts = _range_slice(
        slice(250_000, 500_000, 250_000)
    )  # TODO: For larger-memory machines, we could use a larger max here, e.g. 5M

    setup_params = itertools.product(
        rng_seeds,
        left_key_unique_counts,
        right_key_unique_counts,
        left_rowcounts,
        right_rowcounts,
    )

    # Trial parameters.
    hows = [
        "left",
        "right",
        "inner",
    ]  # TODO: Add 'outer'; skip 'right', that's the same logic as left
    left_keeps = [None, "first", "last"]
    right_keeps = [None, "first", "last"]

    trial_params = lambda: itertools.product(hows, left_keeps, right_keeps)

    # Datasets containing timing data and parameters from the trials in this benchmark.
    benchmark_data: List[Dataset] = []

    for (
        rng_seed,
        left_key_unique_count,
        right_key_unique_count,
        left_rowcount,
        right_rowcount,
    ) in setup_params:
        #
        # Setup phase. The data here is used for both the warmup and the real, timed function invocations.
        #
        # Make sure to re-initialize the RNG each time so we get a repeatable result.
        rng = default_rng(rng_seed)

        left_ds = rand_dataset(left_rowcount, rng, left_key_unique_count)
        right_ds = rand_dataset(right_rowcount, rng, right_key_unique_count)

        # Sweep over trial parameters; all trials sharing the same setup parameters
        # share the same data created from those setup parameters.
        for how, keep_left, keep_right in trial_params():
            # Allocate an array to hold the raw timing data.
            # TODO: Change to use TimeSpan?
            timing_data = empty(iters, dtype=np.int64)

            for is_warmup in (True, False):
                loop_count = warmup_iters if is_warmup else iters
                logger.info(
                    f"bench_merge2:\tmode={'warmup' if is_warmup else 'bench'}\tloops={loop_count}\tleft_rowcount={left_rowcount}\tright_rowcount={right_rowcount}\thow={how}\tkeep_left={keep_left}\tkeep_right={keep_right}"
                )

                for i in range(loop_count):
                    start_time_ns = timestamper()

                    ### The actual function invocation ###
                    # TODO: This could be passed in as a lambda, and we call it with the data constructed
                    #  in the setup phase + any "parameters" that weren't used in the setup phase; or for
                    #  simplicity just pass all the parameters + the data constructed in the setup phase,
                    #  the lambda doesn't need to use all of it.
                    keep = keep_left, keep_right
                    merge2(
                        left_ds,
                        right_ds,
                        on="key",
                        how=how,
                        keep=keep,
                        suffixes=("_x", "_y"),
                    )

                    ### Store the timing results (if this was a real invocation).
                    call_nanos = timestamper() - start_time_ns
                    if not is_warmup:
                        timing_data[i] = call_nanos

            # Create a mini Dataset with the timing results for this run.
            # Capture the timing results along with the other options used for the function invocations.
            trial_data = create_trial_dataset(
                timing_data,
                {
                    # Setup parameters
                    "rng_seed": rng_seed,
                    "left_key_unique_count": left_key_unique_count,
                    "right_key_unique_count": right_key_unique_count,
                    "left_rowcount": left_rowcount,
                    "right_rowcount": right_rowcount,
                    # Trial parameters
                    "how": how,
                    "keep_left": "" if keep_left is None else keep_left,
                    "keep_right": "" if keep_right is None else keep_right,
                },
            )
            benchmark_data.append(trial_data)

    # hstack all of the individual Datasets together into one large Dataset and return it.
    return Dataset.hstack(benchmark_data, destroy=True)


def bench_merge2_with_settings():
    # Define the dimensions (and their values) we want to sweep over for the benchmarking
    how_options = [
        "left",
        "right",
        "inner",
    ]  # TODO: Add 'outer'; skip 'right', that's the same logic as left
    keeps = [None, "first", "last"]
    keep_options = list(product(keeps, keeps))
    left_datasets, right_datasets = generate_merge_datasets()
    warmup_iters, benchmark_iters = 1, 5

    @benchmark(
        benchmark_params={
            # _merge2 parameters
            "left_ds": left_datasets,
            "right_ds": right_datasets,
            "on": ["key"],
            "how": how_options,
            "keeps": keep_options,
            # riptable settings
            "thread_count": [1, 2, 4, 8],
            "recycler": [True, False],
        },
        # benchmark options
        warmup_iterations=warmup_iters,
        benchmark_iterations=benchmark_iters,
        enable_bench_estimators=True,  # TODO
    )
    def bench_merge2(left_ds, right_ds, on, how, keeps):
        merge2(left_ds, right_ds, on=on, how=how, keep=keeps, suffixes=("_x", "_y"))

    return bench_merge2()


def bench_merge_pandas(**kwargs) -> Dataset:
    """Merge benchmark for pandas."""
    import pandas as pd

    # Fixed parameters which apply to all of the trials in this benchmark.
    warmup_iters = 1
    iters = 5

    # Setup parameters.
    rng_seeds = [12345]
    left_key_unique_counts = [100]
    right_key_unique_counts = [10000]
    left_rowcounts = _range_slice(
        slice(500_000, 2_000_000, 500_000)
    )  # TODO: For larger-memory machines, we could use a larger max here, e.g. 7M
    right_rowcounts = _range_slice(
        slice(250_000, 500_000, 250_000)
    )  # TODO: For larger-memory machines, we could use a larger max here, e.g. 5M

    setup_params = itertools.product(
        rng_seeds,
        left_key_unique_counts,
        right_key_unique_counts,
        left_rowcounts,
        right_rowcounts,
    )

    # Trial parameters.
    hows = [
        "left",
        "right",
        "inner",
    ]  # TODO: Add 'outer' once we add it for merge and merge2.
    left_keeps = [None, "first", "last"]
    right_keeps = [None, "first", "last"]

    trial_params = lambda: itertools.product(hows, left_keeps, right_keeps)

    # Datasets containing timing data and parameters from the trials in this benchmark.
    benchmark_data: List[Dataset] = []

    for (
        rng_seed,
        left_key_unique_count,
        right_key_unique_count,
        left_rowcount,
        right_rowcount,
    ) in setup_params:
        #
        # Setup phase. The data here is used for both the warmup and the real, timed function invocations.
        #
        # Make sure to re-initialize the RNG each time so we get a repeatable result.
        rng = default_rng(rng_seed)

        left_df = rand_dataset(left_rowcount, rng, left_key_unique_count).to_pandas(use_nullable=True)
        right_df = rand_dataset(right_rowcount, rng, right_key_unique_count).to_pandas(use_nullable=True)

        # Sweep over trial parameters; all trials sharing the same setup parameters
        # share the same data created from those setup parameters.
        for how, keep_left, keep_right in trial_params():
            # Allocate an array to hold the raw timing data.
            # TODO: Change to use TimeSpan?
            timing_data = empty(iters, dtype=np.int64)

            for is_warmup in (True, False):
                loop_count = warmup_iters if is_warmup else iters

                for i in range(loop_count):
                    start_time_ns = timestamper()

                    ### The actual function invocation ###
                    # Pandas merge doesn't support 'keep' (yet: see https://github.com/pandas-dev/pandas/issues/31332 ),
                    # so we need to do it manually with drop_duplicates first.
                    if keep_left is not None:
                        left_df = left_df.drop_duplicates(subset="key", keep=keep_left)
                    if keep_right is not None:
                        right_df = right_df.drop_duplicates(subset="key", keep=keep_right)

                    left_df.merge(right_df, on="key", how=how)
                    ### Actual function invocation ends here ###

                    ### Store the timing results (if this was a real invocation).
                    call_nanos = timestamper() - start_time_ns
                    if not is_warmup:
                        timing_data[i] = call_nanos

            # Create a mini Dataset with the timing results for this run.
            # Capture the timing results along with the other options used for the function invocations.
            trial_data = create_trial_dataset(
                timing_data,
                {
                    # Setup parameters
                    "rng_seed": rng_seed,
                    "left_key_unique_count": left_key_unique_count,
                    "right_key_unique_count": right_key_unique_count,
                    "left_rowcount": left_rowcount,
                    "right_rowcount": right_rowcount,
                    # Trial parameters
                    "how": how,
                    "keep_left": "" if keep_left is None else keep_left,
                    "keep_right": "" if keep_right is None else keep_right,
                },
            )
            benchmark_data.append(trial_data)

    # hstack all of the individual Datasets together into one large Dataset and return it.
    return Dataset.hstack(benchmark_data, destroy=True)


def compare_merge(**kwargs) -> Dataset:
    """Run all 'merge' benchmarks and return the combined results."""

    # Run the benchmark for rt.merge.
    rtmerge_bench_data = bench_merge(**kwargs)

    # Using the 'how' column and our knowledge of how rt.merge operates,
    # synthesize 'keep_left' and 'keep_right' Categoricals for the rt.merge
    # results so they can be better compared to the other implementations.
    # TODO: This could be removed in the future once we support jagged/non-rectangular
    #       benchmark parameters, since we can pass these in to the benchmark
    #       and have them recorded in the usual way.
    rtmerge_bench_data["keep_left"] = Categorical(
        rtmerge_bench_data["how"].map({"right": "last", "inner": "last", "outer": "last"}, invalid="")
    )
    rtmerge_bench_data["keep_right"] = Categorical(
        rtmerge_bench_data["how"].map({"left": "last", "inner": "last", "outer": "last"}, invalid="")
    )

    # Run the benchmark for rt.merge2.
    rtmerge2_bench_data = bench_merge2(**kwargs)

    # Run the benchmark for pandas.
    pdmerge_bench_data = bench_merge_pandas(**kwargs)

    # Combine the results for the different implementations into
    # a single Dataset and return it.
    return create_comparison_dataset(
        {
            "rt_merge": rtmerge_bench_data,
            "rt_merge2": rtmerge2_bench_data,
            "pandas": pdmerge_bench_data,
        },
        destroy=True,
    )



"""
Utility function for rt.
These functions (may) have dependence on additional libraries and therefore should
_NOT_ be imported in __init__.py or any other such core (like rt_appconfig.py).
"""

__all__ = [
    "dataset_as_pandas_df",
    "dataset_from_pandas_df",
]

import warnings

from ..rt_dataset import Dataset


def dataset_from_pandas_df(df, tz="UTC"):
    """
    This function is deprecated, please use riptable.Dataset.from_pandas.

    Creates a riptable Dataset from a pandas DataFrame. Pandas categoricals
    and datetime arrays are converted to their riptable counterparts.
    Any timezone-unaware datetime arrays (or those using a timezone not
    recognized by riptable) are localized to the timezone specified by the
    tz parameter.

    Recognized pandas timezones:
        UTC, GMT, US/Eastern, and Europe/Dublin

    Parameters
    ----------
    df: DataFrame
        The pandas DataFrame to be converted
    tz: string
        A riptable-supported timezone ('UTC', 'NYC', 'DUBLIN', 'GMT')

    Returns
    -------
    Dataset

    See Also
    --------
    riptable.Dataset.from_pandas
    riptable.Dataset.to_pandas
    """
    warnings.warn(
        "dataset_from_pandas_df is deprecated and will be removed in future release, "
        "please use riptable.Dataset.from_pandas method",
        FutureWarning,
        stacklevel=2,
    )
    return Dataset.from_pandas(df, tz)


def dataset_as_pandas_df(ds):
    """
    This function is deprecated, please use riptable.Dataset.as_pandas_df method.

    Create a pandas DataFrame from a riptable Dataset.
    Will attempt to preserve single-key categoricals, otherwise will appear as
    an index array. Any bytestrings will be converted to unicode.

    Parameters
    ----------
    ds : Dataset
        The riptable Dataset to be converted.

    Returns
    -------
    DataFrame

    See Also
    --------
    riptable.Dataset.to_pandas
    """
    warnings.warn(
        "dataset_as_pandas_df is deprecated and will be removed in future release, "
        "please use riptable.Dataset.to_pandas method",
        FutureWarning,
        stacklevel=2,
    )
    return ds.to_pandas()



import unittest
from math import nan
from typing import List

import hypothesis
import numpy as np
import pytest
from hypothesis import HealthCheck, assume, event, example, given
from hypothesis.extra.numpy import (
    arrays,
    boolean_dtypes,
    floating_dtypes,
    integer_dtypes,
    unsigned_integer_dtypes,
)
from hypothesis.strategies import one_of
from numpy.testing import (
    assert_allclose,
    assert_almost_equal,
    assert_array_almost_equal,
    assert_array_equal,
    assert_equal,
)

import riptable as rt
from riptable import FA, FastArray
from riptable.Utils.teamcity_helper import is_running_in_teamcity

# riptable custom Hypothesis strategies
from .strategies.helper_strategies import (
    generate_list_ndarrays,
    generate_lists,
    generate_ndarrays,
    ndarray_shape_strategy,
    one_darray_shape_strategy,
)


class TestCat2Keys:
    @pytest.mark.skipif(is_running_in_teamcity(), reason="Hypothesis generation taking too long.")
    @given(keys=generate_list_ndarrays())
    def test_cat2keys_nested_array(self, keys):
        key1, key2 = keys
        multi_cat = rt.cat2keys(key1, key2)
        print(f"key1 {repr(key1)}\nkey2: {repr(key2)}")

    @pytest.mark.skipif(is_running_in_teamcity(), reason="Hypothesis generation taking too long.")
    @given(keys=one_of(generate_lists(), generate_ndarrays()))
    def test_cat2keys(self, keys):
        key1, key2 = keys

        multi_cat = rt.cat2keys(key1, key2)
        assert len(key1) == len(key2)  # add test to check different length lists

        # these are the expected entries in the multi key categorical dictionary
        n = len(key1)
        expected_key1 = set(rt.FastArray([k for _ in range(n) for k in key1]))
        expected_key2 = set(rt.FastArray([k for k in key2 for _ in range(n)]))

        key_itr = iter(multi_cat.category_dict)
        actual_key1 = set(multi_cat.category_dict[next(key_itr)])
        actual_key2 = set(multi_cat.category_dict[next(key_itr)])

        not_nan = lambda x: not np.isnan(x)
        assert not set(
            filter(not_nan, expected_key1.symmetric_difference(actual_key1))
        ), f"\nexpected {expected_key1}\nactual {actual_key1}"
        assert not set(
            filter(not_nan, expected_key2.symmetric_difference(actual_key2))
        ), f"\nexpected {expected_key2}\nactual {actual_key2}"

        # need to handle tuple ordering and string dtype discrepancies
        # Taking the entries one by one of expected_key1 and expected_key2 should produce the
        # cartesian product of key1 and key2.
        # expected_product = {(k1, k2) for k1, k2 in itertools.product(key1, key2)}
        # actual_product = {(k1, k2) for k1, k2 in zip(actual_key1, actual_key2)}

        # not_nan = lambda tup: not np.isnan(tup[0]) or not np.isnan(tup[1])
        # assert not set(filter(not_nan, expected_product.symmetric_difference(actual_product))), f"expected {expected_product}\nactual {actual_product}\nmulti_cat {self.print_cat(multi_cat)}\n{self.print_keys(keys)}"



import itertools
import operator
import os
from typing import Tuple, Union

import numpy
import numpy as np
import numba

import pytest
from numpy.testing import assert_array_compare, assert_array_equal

# Utilize pandas' test data
# from pandas.tests.reshape.merge.test_merge_asof import TestAsOfMerge as TestPandasAsOfMerge

import riptable as rt
from riptable.rt_merge_asof import merge_asof2
from riptable.testing.array_assert import assert_array_or_cat_equal

# TODO: Implement tests for merge_asof2 that check it's behavior when the "on" and/or "by" keycols from 'left' and 'right' have different dtypes.
#         * What's the type of the keycol(s) in the merged Dataset?

# TODO: Implement tests for merge_asof2 to check how it behaves when the "by" keycols from 'left' and/or 'right' have a Grouping
#       containing groups with 0 elements, such as with a filtered Categorical.


def check_merge_asof(ds, ds1, ds2):
    # print(ds)
    assert isinstance(ds, rt.Dataset)
    assert ds.shape[0] == ds1.shape[0]
    assert ds.shape[1] == ds1.shape[1] + ds2.shape[1]
    assert_array_equal(ds.A._np, ds1.A._np)
    assert_array_equal(ds.left_val._np, ds1.left_val._np)
    assert_array_equal(ds.right_val._np, ds.X._np)


class TestMergeAsof2:
    #    def __init__(self):
    #        # Import some pandas pytest fixtures to utilize their test data.
    #        self.trades = TestPandasAsOfMerge.trades
    #        self.quotes = TestPandasAsOfMerge.quotes
    #        self.asof = TestPandasAsOfMerge.asof
    #        self.tolerance = TestPandasAsOfMerge.tolerance
    #        self.allow_exact_matches_data = TestPandasAsOfMerge.allow_exact_matches
    #        self.allow_exact_matches_and_tolerance_data = TestPandasAsOfMerge.allow_exact_matches_and_tolerance

    @pytest.mark.parametrize(
        ("time_dtypes", "tolerance"),
        [
            pytest.param((np.int8, np.int16), None),
            pytest.param((np.int8, np.int16), 10),
            pytest.param((np.float64, np.float32), None),
            pytest.param((np.float64, np.float32), 10.0),
        ],
    )
    def test_merge_asof_backward_exact(self, time_dtypes, tolerance) -> None:
        # Unpack dtypes for the 'on' columns.
        ds1_time_dtype, ds2_time_dtype = time_dtypes

        ds1 = rt.Dataset()
        ds1["Time"] = rt.FA([-3, 0, 1, 4, 6, 8, 9, 11, 16, 19, 30], dtype=ds1_time_dtype)
        ds1["Px"] = rt.FA([8, 10, 12, 15, 11, 10, 9, 13, 7, 9, 10], dtype=np.int16)

        ds2 = rt.Dataset()
        ds2["Time"] = rt.FA([0, 0, 5, 7, 8, 8, 10, 10, 12, 15, 17, 20], dtype=ds2_time_dtype)
        ds2["Vols"] = rt.FA([0, 20, 22, 19, 3, 24, 5, 25, 26, 27, 28, 29])

        target = rt.Dataset()
        target.Time = ds1.Time
        target.Px = ds1.Px
        target.Vols = rt.FA([ds2["Vols"].inv, 20, 20, 20, 22, 24, 24, 25, 27, 28, 29], dtype=ds2["Vols"].dtype)
        # For seeing which time was matched against (from the right side) for each row
        target.Time_y = rt.FA([ds2["Time"].inv, 0, 0, 0, 5, 8, 8, 10, 15, 17, 20], dtype=ds2["Time"].dtype)

        merged = merge_asof2(ds1, ds2, on="Time", direction="backward", matched_on="Time_y", tolerance=tolerance)

        for key in merged.keys():
            assert_array_or_cat_equal(
                merged[key], target[key], err_msg=f"Column '{key}' differs between the actual and expected."
            )

    def test_merge_asof_backward_noexact(self) -> None:
        ds1 = rt.Dataset()
        ds1["Time"] = rt.FA([0, 1, 4, 6, 8, 9, 11, 16, 19, 30], dtype=np.int8)
        ds1["Px"] = rt.FA([10, 12, 15, 11, 10, 9, 13, 7, 9, 10], dtype=np.int16)

        ds2 = rt.Dataset()
        ds2["Time"] = rt.FA([0, 0, 5, 7, 8, 8, 10, 10, 12, 15, 17, 20], dtype=np.int16)
        ds2["Vols"] = rt.FA([0, 20, 22, 19, 3, 24, 5, 25, 26, 27, 28, 29])

        target = rt.Dataset()
        target.Time = ds1.Time
        target.Px = ds1.Px
        target.Vols = rt.FA([ds2["Vols"].inv, 20, 20, 22, 19, 24, 25, 27, 28, 29])
        # For seeing which time was matched against (from the right side) for each row
        target.Time_y = rt.FA([ds2["Time"].inv, 0, 0, 5, 7, 8, 10, 15, 17, 20], dtype=ds2["Time"].dtype)

        merged = merge_asof2(ds1, ds2, on="Time", direction="backward", allow_exact_matches=False, matched_on="Time_y")

        for key in merged.keys():
            assert_array_or_cat_equal(
                merged[key], target[key], err_msg=f"Column '{key}' differs between the actual and expected."
            )

    def test_merge_asof_backward_grouped_exact(self):
        ds1 = rt.Dataset({"A": [1, 5, 10], "left_val": ["a", "b", "c"], "left_grp": [1, 1, 1]})
        ds2 = rt.Dataset(
            {
                "X": [1, 2, 3, 6, 7],
                "right_val": [1, 2, 3, 6, 7],
                "right_grp": [1, 1, 1, 1, 1],
            }
        )

        ds = merge_asof2(
            ds1,
            ds2,
            left_on="A",
            right_on="X",
            left_by="left_grp",
            right_by="right_grp",
            direction="backward",
        )
        check_merge_asof(ds, ds1, ds2)
        assert_array_equal(ds.left_grp._np, ds.right_grp._np)
        assert_array_compare(operator.__ge__, ds.A, ds.X)

    def test_merge_asof_backward_grouped_noexact(self) -> None:
        ds1 = rt.Dataset({"A": [1, 5, 10], "left_val": ["a", "b", "c"], "left_grp": [1, 1, 1]})
        ds2 = rt.Dataset(
            {
                "X": [1, 2, 3, 6, 7],
                "right_val": [1, 2, 3, 6, 7],
                "right_grp": [1, 1, 1, 1, 1],
            }
        )

        ds = merge_asof2(
            ds1,
            ds2,
            left_on="A",
            right_on="X",
            left_by="left_grp",
            right_by="right_grp",
            allow_exact_matches=False,
        )
        check_merge_asof(ds, ds1, ds2)
        assert_array_equal(ds.left_grp._np[1:], ds.right_grp._np[1:])
        assert_array_compare(operator.__gt__, ds.A[1:], ds.X[1:])

    @pytest.mark.parametrize("time_multiplier", [1, 2])
    def test_merge_asof_forward_exact(self, time_multiplier: int) -> None:
        ds1 = rt.Dataset()
        ds1["Time"] = rt.FA([-3, 0, 1, 4, 6, 8, 9, 11, 16, 19, 30], dtype=np.int8)
        ds1["Px"] = rt.FA([8, 10, 12, 15, 11, 10, 9, 13, 7, 9, 10], dtype=np.int16)

        ds2 = rt.Dataset()
        end_time = 20 * time_multiplier
        ds2["Time"] = rt.FA([0, 0, 5, 7, 8, 8, 10, 10, 12, 15, 17, end_time, end_time, end_time], dtype=np.int16)
        ds2["Vols"] = rt.FA([0, 20, 22, 19, 3, 24, 5, 25, 26, 27, 28, 29, 21, 33])

        target = rt.Dataset()
        target.Time = ds1.Time
        target.Px = ds1.Px
        target.Vols = rt.FA(
            [0, 0, 22, 22, 19, 3, 5, 26, 28, 29, (ds2["Vols"].inv if time_multiplier == 1 else 29)],
            dtype=ds2["Vols"].dtype,
        )
        # For seeing which time was matched against (from the right side) for each row
        target.Time_y = rt.FA(
            [0, 0, 5, 5, 7, 8, 10, 12, 17, end_time, (ds2["Time"].inv if time_multiplier == 1 else end_time)],
            dtype=ds2["Time"].dtype,
        )

        merged = merge_asof2(ds1, ds2, on="Time", direction="forward", matched_on="Time_y")

        for key in merged.keys():
            assert_array_or_cat_equal(
                merged[key], target[key], err_msg=f"Column '{key}' differs between the actual and expected."
            )

    def test_merge_asof_forward_noexact(self) -> None:
        ds1 = rt.Dataset()
        ds1["Time"] = rt.FA([0, 1, 4, 6, 8, 9, 11, 16, 19, 30], dtype=np.int8)
        ds1["Px"] = rt.FA([10, 12, 15, 11, 10, 9, 13, 7, 9, 10], dtype=np.int16)

        ds2 = rt.Dataset()
        ds2["Time"] = rt.FA([0, 0, 5, 7, 8, 8, 10, 10, 12, 15, 17, 20], dtype=np.int16)
        ds2["Vols"] = rt.FA([0, 20, 22, 19, 3, 24, 5, 25, 26, 27, 28, 29])

        target = rt.Dataset()
        target.Time = ds1.Time
        target.Px = ds1.Px
        target.Vols = rt.FA([22, 22, 22, 19, 5, 5, 26, 28, 29, ds2["Vols"].inv])
        # For seeing which time was matched against (from the right side) for each row
        target.Time_y = rt.FA([5, 5, 5, 7, 10, 10, 12, 17, 20, ds2["Time"].inv], dtype=ds2["Time"].dtype)

        merged = merge_asof2(ds1, ds2, on="Time", direction="forward", allow_exact_matches=False, matched_on="Time_y")

        for key in merged.keys():
            assert_array_or_cat_equal(
                merged[key], target[key], err_msg=f"Column '{key}' differs between the actual and expected."
            )

    def test_merge_asof_forward_grouped_exact(self) -> None:
        ds1 = rt.Dataset({"A": [1, 5, 10], "left_val": ["a", "b", "c"], "left_grp": [1, 1, 1]})
        ds2 = rt.Dataset(
            {
                "X": [1, 2, 3, 6, 7],
                "right_val": [1, 2, 3, 6, 7],
                "right_grp": [1, 1, 1, 1, 1],
            }
        )

        ds = merge_asof2(
            ds1,
            ds2,
            left_on="A",
            right_on="X",
            left_by="left_grp",
            right_by="right_grp",
            direction="forward",
        )
        check_merge_asof(ds, ds1, ds2)
        assert_array_equal(ds.left_grp._np[:-1], ds.right_grp._np[:-1])
        assert_array_compare(operator.__le__, ds.A[:-1], ds.X[:-1])

    def test_merge_asof_forward_grouped_noexact(self) -> None:
        ds1 = rt.Dataset({"A": [1, 5, 10], "left_val": ["a", "b", "c"], "left_grp": [1, 1, 1]})
        ds2 = rt.Dataset(
            {
                "X": [1, 1, 2, 3, 6, 7],
                "right_val": [1, 0, 2, 3, 6, 7],
                "right_grp": [1, 1, 1, 1, 1, 1],
            }
        )

        ds = merge_asof2(
            ds1,
            ds2,
            left_on="A",
            right_on="X",
            left_by="left_grp",
            right_by="right_grp",
            allow_exact_matches=False,
            direction="forward",
        )
        check_merge_asof(ds, ds1, ds2)
        assert_array_equal(ds.left_grp._np[:-1], ds.right_grp._np[:-1])
        assert_array_compare(operator.__lt__, ds.A[:-1], ds.X[:-1])

    def test_merge_asof_nearest_exact(self):
        ds1 = rt.Dataset()
        ds1["Time"] = rt.FA([0, 1, 3, 7, 8, 9, 11, 16, 19, 30])
        ds1["Px"] = rt.FA([10, 12, 15, 11, 10, 9, 13, 7, 9, 10])

        ds2 = rt.Dataset()
        ds2["Time"] = rt.FA([0, 0, 4, 5, 8, 8, 10, 10, 12, 15, 17, 20])
        ds2["Vols"] = rt.FA([0, 20, 22, 19, 23, 18, 24, 25, 26, 27, 28, 29])
        ds2["idx_right"] = rt.arange(ds2.get_nrows())

        target = rt.Dataset()
        target.Time = ds1.Time
        target.Px = ds1.Px
        # For seeing which time was matched against (from the right side) for each row
        target.idx_right = rt.FA([1, 1, 2, 4, 5, 5, 7, 9, 11, 11])
        target.Time_y = ds2["Time"][target.idx_right]
        target.Vols = ds2["Vols"][target.idx_right]

        merged = merge_asof2(ds1, ds2, on="Time", direction="nearest", matched_on="Time_y")

        for key in merged.keys():
            assert_array_equal(
                merged[key], target[key], err_msg=f"Column '{key}' differs between the actual and expected."
            )

        # ds = merge_asof2(ds1, ds2, left_on='A', right_on='X', left_by='left_grp',
        #                            right_by='right_grp', direction='nearest')
        # check_merge_asof(ds, ds1, ds2)
        # for i in range(0, ds.shape[0]):
        #    diff = abs(ds.A[i] - ds.X[i])
        #    mindiff = min(ds.A - ds.X)
        #    self.assertEqual(diff, mindiff, diff + ", " + mindiff);
        # self.assertTrue((ds.left_grp._np == ds.right_grp._np).all())

    def test_merge_asof_nearest_noexact(self):
        ds1 = rt.Dataset()
        ds1["Time"] = rt.FA([0, 1, 3, 7, 8, 9, 11, 16, 19, 30])
        ds1["Px"] = rt.FA([10, 12, 15, 11, 10, 9, 13, 7, 9, 10])

        ds2 = rt.Dataset()
        ds2["Time"] = rt.FA([0, 0, 4, 5, 8, 8, 10, 10, 12, 15, 17, 20])
        ds2["Vols"] = rt.FA([0, 20, 22, 19, 23, 18, 24, 25, 26, 27, 28, 29])
        ds2["idx_right"] = rt.arange(ds2.get_nrows())

        target = rt.Dataset()
        target.Time = ds1.Time
        target.Px = ds1.Px
        # For seeing which time was matched against (from the right side) for each row
        target.idx_right = rt.FA([2, 1, 2, 4, 6, 5, 7, 9, 11, 11])
        target.Time_y = ds2["Time"][target.idx_right]
        target.Vols = ds2["Vols"][target.idx_right]

        merged = merge_asof2(ds1, ds2, on="Time", allow_exact_matches=False, direction="nearest", matched_on="Time_y")

        for key in merged.keys():
            assert_array_equal(
                merged[key], target[key], err_msg=f"Column '{key}' differs between the actual and expected."
            )

    def test_merge_asof_nearest_grouped_exact(self) -> None:
        ds1 = rt.Dataset({"A": [1, 5, 10], "left_val": ["a", "b", "c"], "left_grp": [1, 1, 1]})
        ds2 = rt.Dataset(
            {
                "X": [1, 2, 3, 6, 7],
                "right_val": [1, 2, 3, 6, 7],
                "right_grp": [1, 1, 1, 1, 1],
            }
        )

        merged = merge_asof2(
            ds1,
            ds2,
            on=("A", "X"),
            by=("left_grp", "right_grp"),
            direction="nearest",
        )
        check_merge_asof(merged, ds1, ds2)

        target = ds1.copy(deep=True)
        target.X = rt.FA([1, 6, 7])
        target.right_val = rt.FA([1, 6, 7])
        target.right_grp = rt.FA([1, 1, 1])

        for key in merged.keys():
            assert_array_equal(
                merged[key], target[key], err_msg=f"Column '{key}' differs between the actual and expected."
            )

    def test_merge_asof_nearest_grouped_noexact(self) -> None:
        ds1 = rt.Dataset({"A": [1, 5, 10], "left_val": ["a", "b", "c"], "left_grp": [1, 1, 1]})
        ds2 = rt.Dataset(
            {
                "X": [1, 1, 2, 3, 6, 7],
                "right_val": [1, 0, 2, 3, 6, 7],
                "right_grp": [1, 1, 1, 1, 1, 1],
            }
        )

        merged = merge_asof2(
            ds1,
            ds2,
            on=("A", "X"),
            by=("left_grp", "right_grp"),
            allow_exact_matches=False,
            direction="nearest",
        )
        check_merge_asof(merged, ds1, ds2)

        target = ds1.copy(deep=True)
        target.X = rt.FA([2, 6, 7])
        target.right_val = rt.FA([2, 6, 7])
        target.right_grp = rt.FA([1, 1, 1])

        for key in merged.keys():
            assert_array_equal(
                merged[key], target[key], err_msg=f"Column '{key}' differs between the actual and expected."
            )

    def test_merge_asof_categorical_and_string_keys(self):
        ds1 = rt.Dataset()
        ds1["Time"] = [0, 1, 4, 6, 8, 9, 11, 16, 19, 30]
        ds1["Px"] = [10, 12, 15, 11, 10, 9, 13, 7, 9, 10]

        ds2 = rt.Dataset()
        ds2["Time"] = [0, 0, 5, 7, 8, 10, 12, 15, 17, 20]
        ds2["Vols"] = [20, 21, 22, 23, 24, 25, 26, 27, 28, 29]

        target = rt.Dataset()
        target.Time = ds1.Time
        target.Px = ds1.Px
        target.Vols = [20, 20, 20, 22, 24, 24, 24, 26, 28, 28]
        target.matched_on = [
            0,
            0,
            0,
            5,
            8,
            8,
            8,
            12,
            17,
            17,
        ]  # For seeing which time was matched against (from the right side) for each row

        # Categorical keys
        ds1["Ticker"] = rt.Categorical(["Test"] * 10)
        ds2["Ticker"] = rt.Categorical(["Test", "Blah"] * 5)
        target.Ticker = ds1.Ticker

        ds = merge_asof2(ds1, ds2, on="Time", by="Ticker", matched_on=True)

        for key in ds.keys():
            # TODO: Switch to assert_array_equal here once a string-based Categorical allows np.inf to be used in an equality check
            # assert_array_or_cat_equal(ds[key], target[key], err_msg=f"Column '{key}' differs between the actual and expected.")
            assert (ds[key] == target[key]).all()

        # String keys
        ds1["Ticker"] = rt.FastArray(["Test"] * 10)
        ds2["Ticker"] = rt.FastArray(["Test", "Blah"] * 5)
        target.Ticker = ds1.Ticker

        ds = merge_asof2(ds1, ds2, on="Time", by="Ticker")

        for key in ds.keys():
            assert_array_equal(ds[key], target[key], err_msg=f"Column '{key}' differs between the actual and expected.")

    @pytest.mark.parametrize(
        "left,right,on,should_fail",
        [
            pytest.param(
                rt.Dataset({"A": [1, 5, 10], "left_val": ["a", "b", "c"], "left_grp": [1, 1, 1]}),
                rt.Dataset(
                    {
                        "X": [1, 2, 3, 6, 7],
                        "right_val": [1, 2, 3, 6, 7],
                        "right_grp": [1, 1, 1, 1, 1],
                    }
                ),
                ("A", "X"),
                False,
                id="sorted,sorted",
            ),
            pytest.param(
                rt.Dataset({"A": [1, 10, 5], "left_val": ["a", "b", "c"], "left_grp": [1, 1, 1]}),
                rt.Dataset(
                    {
                        "X": [1, 2, 3, 6, 7],
                        "right_val": [1, 2, 3, 6, 7],
                        "right_grp": [1, 1, 1, 1, 1],
                    }
                ),
                ("A", "X"),
                True,
                id="unsorted,sorted",
            ),
            pytest.param(
                rt.Dataset({"A": [1, 5, 10], "left_val": ["a", "b", "c"], "left_grp": [1, 1, 1]}),
                rt.Dataset(
                    {
                        "X": [1, 2, 3, 6, 4],
                        "right_val": [1, 2, 3, 6, 7],
                        "right_grp": [1, 1, 1, 1, 1],
                    }
                ),
                ("A", "X"),
                True,
                id="sorted,unsorted",
            ),
        ],
    )
    def test_merge_asof_unsorted_fails(
        self,
        left: rt.Dataset,
        right: rt.Dataset,
        on: Union[str, Tuple[str, str]],
        should_fail: bool,
    ):
        """
        Test for verifying 'merge_asof' checks the sortedness of the 'on' columns before merging.
        """
        # Unpack the 'on' parameter (if needed) until 'merge_asof' has been
        # modified to allow a tuple to be passed for the 'on' parameter (instead of left_on+right_on).
        left_on, right_on = on if isinstance(on, tuple) else (on, on)

        if should_fail:
            with pytest.raises(ValueError):
                merge_asof2(
                    left,
                    right,
                    left_on=left_on,
                    right_on=right_on,
                )
        else:
            # Just check that this succeeds without raising any errors;
            # we otherwise don't care about the result of the operation.
            merge_asof2(
                left,
                right,
                left_on=left_on,
                right_on=right_on,
            )

    def test_merge_asof_backward_multiple_by(self) -> None:
        # Adapted from https://github.com/pandas-dev/pandas/blob/32b4222050c7d5eedfa8fa47d5a45fb1d6c966a5/pandas/tests/reshape/merge/test_merge_asof.py#L286
        trades = rt.Dataset(
            {
                "time": rt.DateTimeNano(
                    [
                        "20160525 13:30:00.023",
                        "20160525 13:30:00.023",
                        "20160525 13:30:00.046",
                        "20160525 13:30:00.048",
                        "20160525 13:30:00.050",
                    ],
                    from_tz="America/New_York",
                ),
                "ticker": ["MSFT", "MSFT", "GOOG", "GOOG", "AAPL"],
                "exch": ["ARCA", "NSDQ", "NSDQ", "BATS", "NSDQ"],
                "price": [51.95, 51.95, 720.77, 720.92, 98.00],
                "quantity": [75, 155, 100, 100, 100],
            }
        )

        quotes = rt.Dataset(
            {
                "time": rt.DateTimeNano(
                    [
                        "20160525 13:30:00.023",
                        "20160525 13:30:00.023",
                        "20160525 13:30:00.030",
                        "20160525 13:30:00.041",
                        "20160525 13:30:00.045",
                        "20160525 13:30:00.049",
                    ],
                    from_tz="America/New_York",
                ),
                "ticker": ["GOOG", "MSFT", "MSFT", "MSFT", "GOOG", "AAPL"],
                "exch": ["BATS", "NSDQ", "ARCA", "ARCA", "NSDQ", "ARCA"],
                "bid": [720.51, 51.95, 51.97, 51.99, 720.50, 97.99],
                "ask": [720.92, 51.96, 51.98, 52.00, 720.93, 98.01],
            }
        )

        expected = rt.Dataset(
            {
                "time": rt.DateTimeNano(
                    [
                        "20160525 13:30:00.023",
                        "20160525 13:30:00.023",
                        "20160525 13:30:00.046",
                        "20160525 13:30:00.048",
                        "20160525 13:30:00.050",
                    ],
                    from_tz="America/New_York",
                ),
                "ticker": ["MSFT", "MSFT", "GOOG", "GOOG", "AAPL"],
                "exch": ["ARCA", "NSDQ", "NSDQ", "BATS", "NSDQ"],
                "price": [51.95, 51.95, 720.77, 720.92, 98.00],
                "quantity": [75, 155, 100, 100, 100],
                "bid": [np.nan, 51.95, 720.50, 720.51, np.nan],
                "ask": [np.nan, 51.96, 720.93, 720.92, np.nan],
            }
        )

        merged = merge_asof2(trades, quotes, on="time", by=["ticker", "exch"])

        for key in merged.keys():
            assert_array_or_cat_equal(
                merged[key], expected[key], err_msg=f"Column '{key}' differs between the actual and expected."
            )

    @pytest.mark.parametrize(
        (
            "time_dtypes",
            "tolerance",
        ),
        [
            pytest.param(
                (np.uint8, np.uint16),
                rt.uint16.inv,
                id="Unsigned integer invalid",
                marks=[pytest.mark.skip("Unclear what the expected behavior should be in this case.")],
            ),
            pytest.param((np.uint8, np.int16), -10, id="Negative integer"),
            pytest.param((np.float64, np.float32), np.nan, id="NaN floating point"),
            pytest.param((np.float64, np.float32), -10.0, id="Negative floating point"),
        ],
    )
    def test_merge_asof_backward_bad_tolerance(self, time_dtypes, tolerance) -> None:
        """Test that merge_asof2 rejects invalid/unusable 'tolerance' values."""
        # Unpack dtypes for the 'on' columns.
        ds1_time_dtype, ds2_time_dtype = time_dtypes

        ds1 = rt.Dataset()
        ds1["Time"] = rt.FA([0, 1, 4, 6, 8, 9, 11, 16, 19, 30], dtype=ds1_time_dtype)
        ds1["Px"] = rt.FA([10, 12, 15, 11, 10, 9, 13, 7, 9, 10], dtype=np.int16)

        ds2 = rt.Dataset()
        ds2["Time"] = rt.FA([0, 0, 5, 7, 8, 8, 10, 10, 12, 15, 17, 20], dtype=ds2_time_dtype)
        ds2["Vols"] = rt.FA([0, 20, 22, 19, 3, 24, 5, 25, 26, 27, 28, 29])

        target = rt.Dataset()
        target.Time = ds1.Time
        target.Px = ds1.Px
        target.Vols = rt.FA([20, 20, 20, 22, 24, 24, 25, 27, 28, 29])
        # For seeing which time was matched against (from the right side) for each row
        target.Time_y = rt.FA([0, 0, 0, 5, 8, 8, 10, 15, 17, 20], dtype=ds2["Time"].dtype)

        with pytest.raises(ValueError):
            _ = merge_asof2(ds1, ds2, on="Time", direction="backward", matched_on="Time_y", tolerance=tolerance)



import unittest
from io import StringIO

import numpy as np
from numpy import nan
from numpy.testing import assert_array_equal, assert_equal

from riptable import FastArray as FA
from riptable import mask_and, mask_andnot, mask_or, mask_xor

from .utils import redirectStdoutCtx

try:
    # bottleneck optional
    import bottleneck as bn

    class FastArray_BN_Test(unittest.TestCase):
        def test_masking(self):
            np_bools = [np.random.randint(low=0, high=2, size=10_000_000).astype(bool) for i in range(6)]

            bool_length = len(np_bools)
            result = np_bools[0] + np_bools[1]
            for i in range(2, bool_length):
                result += np_bools[i]

            result2 = mask_or(np_bools)
            assert_array_equal(result, result2)

            result = np_bools[0] * np_bools[1]
            for i in range(2, bool_length):
                result *= np_bools[i]

            result2 = mask_and(np_bools)
            assert_array_equal(result, result2)

            result = np_bools[0] ^ np_bools[1]
            for i in range(2, bool_length):
                result ^= np_bools[i]

            result2 = mask_xor(np_bools)
            assert_array_equal(result, result2)

            result3 = mask_xor(result2, result, result2, result)
            assert_equal(result3.sum(), 0)

            result3 = mask_xor(result2, result)
            assert_equal(result3.sum(), 0)

            result3 = mask_xor(result2, result, result2)
            assert_array_equal(result, result3)

            result = np_bools[0] * ~np_bools[1]
            for i in range(2, bool_length):
                result = result * ~np_bools[i]
            result2 = mask_andnot(np_bools)
            assert_array_equal(result, result2)

        def test_noncontig(self):
            # no longer a warning
            a = FA([1, 2, 3, 4, 5, 6])
            b = FA(a[::2])
            assert_array_equal(b, [1, 3, 5])

        def test_map(self):
            a = FA([1, 1, 1, 2, 2, 2])
            d = {1: 10, 2: 20}
            c = a.map(d)
            assert_array_equal(c, [10, 10, 10, 20, 20, 20])

        def test_push(self):
            a = FA([5, nan, nan, 6, nan])
            b = a.push()
            assert_array_equal(b, [5, 5, 5, 6, 6])

        def test_rank(self):
            a = FA([nan, 2, 2, 3])
            b = a.nanrankdata()
            assert_array_equal(b, [nan, 1.5, 1.5, 3.0])
            a = FA([0, 2, 2, 3])
            b = a.rankdata()
            assert_array_equal(b, [1, 2.5, 2.5, 4])

        def test_replace(self):
            a = FA([0, 2, 2, 3])
            b = a.replace(2, 1)
            assert_array_equal(b, [0, 1, 1, 3])

        def test_partition(self):
            a = FA([1, 0, 3, 4, 2])
            b = a.partition2(kth=2)
            assert_array_equal(b, [1, 0, 2, 4, 3])

            a = FA([10, 0, 30, 40, 20])
            b = a.argpartition2(kth=2)
            assert_array_equal(b, [0, 1, 4, 3, 2])

        def test_out(self):
            a = FA(np.arange(100))
            c = a
            b = FA(np.arange(100.0))
            np.add(b, b, out=a)
            assert_array_equal(a, c)
            # this is not supported in numpy and also not in riptable anymore
            # np.floor(b,out=a)
            # assert_array_equal(a.astype(b.dtype),b)

        def test_ledger(self):
            self.maxDiff = None
            sio = StringIO()
            with redirectStdoutCtx(sio):
                FA._V1()
                FA._V2()
                FA._OFF()
                FA._ON()
                FA._TOFF()
                FA._TON()
                FA._LON()
                FA._ROFF()
                FA._RON()
                FA._RDUMP()
                FA._LOFF()
                FA._LDUMP()
                FA._LCLEAR()
                FA._GCNOW()
                FA._GCSET(100)
                FA.Verbose = 3
                FA._V1()
                a = FA([1, 2, 3])
                b = a.copy()
                a += a
                del a
                a = np.arange(10000)
                a = FA(a)
                a = a + a
                a = a + a
                b = a._np
                del b
                del a
                FA._V0()

except:
    pass


if __name__ == "__main__":
    tester = unittest.main()



import os
import unittest

import pytest
from numpy.testing import assert_array_equal

import riptable as rt
from riptable import *
from riptable.rt_enum import GROUPBY_KEY_PREFIX
from riptable.rt_sds import SDSMakeDirsOn

# change to true since we write into /tests directory
SDSMakeDirsOn()

# TODO: Replace these two functions with calls to assert_array_equal to get better diagnostics if the assertion fails.
def arr_all(a):
    return bool(np.all(a))


def arr_eq(a, b):
    return arr_all(a == b)


#
# TODO: Modify the tests in this file to -- at minimum -- generate the random data they use with a deterministic
#       seed, and also regenerate it within each test (to avoid potential issues sharing the data across tests).
#       It would be even better if we could generate the arrays used in the tests with hypothesis, since that'll
#       make it easier to describe the random arrays we want and it'd let us easily reproduce a random test case
#       if we find one that leads to a failure.
#

uniquelen = 7
arrlen = 50
modulo = 3
keyname = "keycol"
d = {"data" + str(i): np.random.rand(arrlen) for i in range(7)}
f = logical(arange(arrlen) % modulo)

randi = np.random.randint(10000, 20000, uniquelen)
randi = np.random.choice(randi, arrlen)
rands = randi.astype("S")
randu = randi.astype("U")
randf = randi.astype(np.float32)


class Grouping_Test(unittest.TestCase):
    def test_head(self):
        pytest.skip("This test is not yet implemented.")

    def test_tail(self):
        pytest.skip("This test is not yet implemented.")

    def test_single_key(self):

        for vals in [randi, rands, randu, randf]:
            vals = vals.view(FA)
            vals.set_name(keyname)
            ds = Dataset({keyname: vals}, unicode=True)
            gb = ds.gb(keyname)
            gb_result = gb.sum(d)

            c = Categorical(vals)
            c_result = c.sum(d)

            ds_c = Dataset({keyname: c}, unicode=True)
            gbc = ds_c.gb(keyname)
            gbc_result = gbc.sum(d)

            self.assertTrue(
                gb_result.equals(c_result),
                msg=f"Groupby and categorical results did not match for single key dtype {vals.dtype}.",
            )
            assert_array_equal(gb_result.label_get_names(), c_result.label_get_names())

            self.assertTrue(
                c_result.equals(gbc_result),
                msg=f"Categorical and groupby categorical results did not match for single key dtype {vals.dtype}.",
            )
            assert_array_equal(c_result.label_get_names(), gbc_result.label_get_names())

    def test_single_key_filter(self):
        for vals in [randi, rands, randu, randf]:
            vals = vals.view(FA)
            vals.set_name(keyname)
            ds = Dataset({keyname: vals}, unicode=True)
            gb = ds.gb(keyname, filter=f)
            gb_result = gb.sum(d, showfilter=True)

            c = Categorical(vals, filter=f)
            c_result = c.sum(d, showfilter=True)

            # reset filter
            c = Categorical(vals)
            ds_c = Dataset({keyname: c}, unicode=True)
            gbc = ds_c.gb(keyname, filter=f)
            gbc_result = gbc.sum(d, showfilter=True)

            self.assertTrue(
                gb_result.equals(c_result),
                msg=f"Groupby and categorical results did not match for single key dtype {vals.dtype}.",
            )
            assert_array_equal(gb_result.label_get_names(), c_result.label_get_names())

            # TJD change
            for colname in c_result:
                if colname != keyname:
                    self.assertTrue(
                        arr_eq(c_result[colname], gbc_result[colname]),
                        msg=f"Categorical and groupby categorical results did not match for single key dtype {vals.dtype}.",
                    )
            self.assertTrue(
                c_result.equals(gbc_result),
                msg=f"Categorical and groupby categorical results did not match for single key dtype {vals.dtype}.",
            )
            assert_array_equal(c_result.label_get_names(), gbc_result.label_get_names())

    def test_multikey(self):
        keynames = ["keycol_0", "keycol_1"]
        for vals1 in [randi, rands, randu, randf]:
            vals1 = vals1.view(FA)
            vals1.set_name(keynames[0])
            for vals2 in [randi, rands, randu, randf]:
                if vals1.dtype == vals2.dtype:
                    continue
                vals2 = vals2.view(FA)
                vals2.set_name(keynames[1])

                vals = [vals1, vals2]

                ds = Dataset(dict(zip(keynames, vals)), unicode=True)
                gb = ds.gb(keynames)
                gb_result = gb.sum(d)

                c = Categorical(vals, sort_gb=True, unicode=True)
                c_result = c.sum(d)

                ds_c = Dataset({"keycol": c}, unicode=True)
                gbc = ds_c.gb("keycol")
                gbc_result = gbc.sum(d)

                assert_array_equal(gb_result.label_get_names(), c_result.label_get_names())
                self.assertTrue(
                    gb_result.equals(c_result),
                    msg=f"Groupby and categorical results did not match for multi key dtype {vals[0].dtype} {vals[1].dtype}.",
                )
                # assert_array_equal(c_result.label_get_names(), gbc_result.label_get_names())
                self.assertTrue(
                    c_result.equals(gbc_result),
                    msg=f"Categorical and groupby categorical results did not match for multi key dtype {c_result} \n {gbc_result}. {c_result} \n {gbc_result}",
                )

    def test_init_paths(self):
        #  empty
        with self.assertRaises(ValueError):
            g = Grouping({})

        # array mismatch
        with self.assertRaises(ValueError):
            g = Grouping({"a": arange(3), "b": arange(4)})

        # non-fastarray
        g = Grouping([np.arange(3), np.arange(3)])
        gdict = g._grouping_dict
        for v in gdict.values():
            self.assertTrue(isinstance(v, FastArray))

        # named arrays
        ds = Dataset({"col1": arange(3), "col2": arange(3)})
        g = Grouping([ds.col1, ds.col2])
        gdict = g._grouping_dict
        for k in gdict:
            self.assertTrue(k in ds)

        # no name arrays
        g = Grouping([arange(3), arange(3)])
        self.assertTrue(
            arr_eq(
                list(g._grouping_dict),
                [GROUPBY_KEY_PREFIX + "_0", GROUPBY_KEY_PREFIX + "_1"],
            )
        )

        # conflicting name arrays
        g = Grouping([ds.col1, ds.col1])
        self.assertTrue(arr_eq(list(g._grouping_dict), ["col1", GROUPBY_KEY_PREFIX + "_c1"]))

    def test_auto_naming(self):
        c = Categorical(["a", "a", "b", "c", "a"])
        self.assertTrue(c.get_name() is None)
        ds = Dataset({"catcol": c})
        result = ds.catcol.sum(arange(5))
        self.assertTrue(arr_eq(["catcol"], result.label_get_names()))

        c = Categorical([FA(["a", "a", "b", "c", "a"]), FA([1, 1, 2, 3, 1])])
        self.assertTrue(c.get_name() is None)
        ds = Dataset({"mkcol": c})
        result = ds.mkcol.sum(arange(5))
        self.assertTrue(arr_eq(["mkcol_0", "mkcol_1"], result.label_get_names()))

        arr1 = FA(["a", "a", "b", "c", "a"])
        arr1.set_name("mystrings")
        arr2 = FA([1, 1, 2, 3, 1])
        arr2.set_name("myints")
        c = Categorical([arr1, arr2])
        ds = Dataset({"mkcol": c})
        result = ds.mkcol.sum(arange(5))
        self.assertTrue(arr_eq(["mystrings", "myints"], result.label_get_names()))

    def test_make_ifirst(self):
        int_types = [np.int8, np.int16, np.int32, np.int64]
        for dt in int_types:
            arr = ones(174_763, dtype=dt)
            ifirst = makeifirst(arr, 1)
            assert ifirst[1] == 0
            assert isnan(ifirst)[0]

            ifirst = makeifirst(arr, 2)
            assert ifirst[1] == 0
            assert isnan(ifirst)[0]
            assert isnan(ifirst)[2]

        c = Categorical(["a", "a", "b", "c", "a"])
        f = FastArray([True, True, False, True, True])
        cfilter = c.set_valid(f)
        ifirst = makeifirst(cfilter, 2)
        assert_array_equal(ifirst[1:], FA([0, 3]))
        assert isnan(ifirst)[0]

        ifirst = makeifirst(c, c.unique_count, filter=f)
        assert isnan(ifirst)[0]
        assert isnan(ifirst)[2]
        assert ifirst[1] == 0
        assert ifirst[-1] == 3

        c = Categorical([1, 1, 1, 1, 2, 1], ["a", "b", "c"])
        ifirst = makeifirst(c, c.unique_count)
        assert_array_equal(ifirst[1:3], FA([0, 4]))
        assert isnan(ifirst)[0]
        assert isnan(ifirst)[-1]

        strs = np.random.choice(arange(100, 200).astype("S"), 1287)

        ds = Dataset({"stringvals": strs})
        dsg = ds.gb("stringvals").grouping

        c = Categorical(ds.stringvals, ordered=False)
        assert_array_equal(makeifirst(c, c.unique_count)[1:], dsg.iFirstKey)
        assert_array_equal(c.grouping.ifirstkey, dsg.ifirstkey)

        c = Categorical(ds.stringvals, ordered=False, base_index=0)
        assert_array_equal(c.grouping.ifirstkey, dsg.ifirstkey)

    def test_makeilast(self):
        int_types = [np.int8, np.int16, np.int32, np.int64]
        for dt in int_types:
            arr = ones(174_763, dtype=dt)
            ilast = makeilast(arr, 1)
            assert ilast[1] == (len(arr) - 1)
            assert isnan(ilast)[0]

            ilast = makeilast(arr, 2)
            assert ilast[1] == (len(arr) - 1)
            assert isnan(ilast)[0]
            assert isnan(ilast)[2]

        c = Categorical(["a", "a", "b", "c", "a"])
        f = FastArray([True, True, False, True, True])
        cfilter = c.set_valid(f)
        ilast = makeilast(cfilter, 2)
        assert_array_equal(ilast[1:], FA([4, 3]))
        assert isnan(ilast)[0]

        ilast = makeilast(c, c.unique_count, filter=f)
        assert isnan(ilast)[0]
        assert isnan(ilast)[2]
        assert ilast[1] == 4
        assert ilast[-1] == 3

        c = Categorical([1, 1, 1, 1, 2, 1], ["a", "b", "c"])
        ilast = makeilast(c, c.unique_count)
        assert_array_equal(ilast[1:3], FA([5, 4]))
        assert isnan(ilast)[0]
        assert isnan(ilast)[-1]

        strs = np.random.choice(arange(100, 200).astype("S"), 1287)

        ds = Dataset({"stringvals": strs})
        dsg = ds.gb("stringvals").grouping

        c = Categorical(ds.stringvals, ordered=False)
        assert_array_equal(makeilast(c, c.unique_count)[1:], dsg.ilastkey)
        assert_array_equal(c.grouping.ilastkey, dsg.ilastkey)

        c = Categorical(ds.stringvals, ordered=False, base_index=0)
        assert_array_equal(c.grouping.ilastkey, dsg.ilastkey)

    def test_isin(self):
        c = Categorical(
            [
                np.random.choice(["aa", "b", "bbb", "c"], 40),
                np.random.choice([1, 2, 3], 40).astype(dtype=np.int8),
            ]
        )
        uniquelist = c.grouping.uniquelist
        tups = [t for t in zip(uniquelist[0].astype("U"), uniquelist[1].astype(np.float32))]
        self.assertTrue(arr_all(c.grouping.isin(tups)))

        c = Categorical(
            [
                np.random.choice(["aa", "b", "bbb", "c"], 40),
                np.random.choice([1, 2, 3], 40).astype(dtype=np.float32),
            ]
        )
        uniquelist = c.grouping.uniquelist
        tups = [t for t in zip(uniquelist[0].astype("U"), uniquelist[1].astype(np.int8))]
        self.assertTrue(arr_all(c.grouping.isin(tups)))

    def test_isin_enum(self):
        c = Categorical([10, 10, 10, 20, 30, 20, 10, 20, 20, 40], {"a": 30, "b": 20, "c": 10})
        ten_mask = FastArray([True, True, True, False, False, False, True, False, False, False])
        self.assertTrue(arr_eq(c.isin(10), ten_mask))
        self.assertTrue(arr_eq(c.isin([10]), ten_mask))
        self.assertTrue(arr_eq(c.isin(FastArray(10)), ten_mask))

        self.assertTrue(arr_eq(c.isin("c"), ten_mask))
        self.assertTrue(arr_eq(c.isin(b"c"), ten_mask))
        self.assertTrue(arr_eq(c.isin(["c"]), ten_mask))
        self.assertTrue(arr_eq(c.isin(np.array(["c"])), ten_mask))
        self.assertTrue(arr_eq(c.isin(FA("c")), ten_mask))

        with self.assertRaises(TypeError):
            _ = c.isin(1.23)

    def test_hstack_base_0(self):
        c7 = Categorical([1, 1], ["a", "b"], base_index=0)
        c8 = Categorical([1, 0], ["b", "c"], base_index=0)
        c9 = hstack([c7, c8])

        self.assertTrue(isinstance(c9, Categorical))
        self.assertEqual(c9.base_index, 0)
        self.assertTrue(arr_eq(c9._fa, FA([1, 1, 2, 1])))
        self.assertTrue(arr_eq(c9.expand_array, FA(["b", "b", "c", "b"])))

    # TODO use pytest fixtures in place of 'paths' and move this and similar test cases into SDS save / load tests
    def test_stack_load_base_0(self):
        c1 = Categorical(np.random.choice(5, 30), ["a", "b", "c", "d", "e"], base_index=0)
        c2 = Categorical(np.random.choice(5, 30), ["e", "f", "g", "a", "b"], base_index=0)

        ds1 = Dataset({"catcol": c1})
        ds2 = Dataset({"catcol": c2})

        paths = [r"riptable/tests/temp/ds1.sds", r"riptable/tests/temp/ds2.sds"]
        ds1.save(paths[0])
        ds2.save(paths[1])

        c_hstack = hstack([c1, c2])
        self.assertTrue(isinstance(c_hstack, Categorical))
        self.assertEqual(c_hstack.base_index, 0)

        pds = load_sds(paths, stack=True)
        self.assertTrue(isinstance(pds, PDataset))
        c = pds.catcol
        self.assertTrue(isinstance(c, Categorical))
        self.assertEqual(c.base_index, 0)
        self.assertTrue(arr_eq(c_hstack, c))

        for p in paths:
            os.remove(p)

    def test_hstack_ordered(self):
        c1 = Categorical([1, 2, 3], ["a", "b", "c"])
        self.assertTrue(c1.ordered)
        c2 = Categorical([1, 2, 3], ["b", "a", "z"])
        self.assertFalse(c2.ordered)

        st = hstack([c1, c2])
        self.assertTrue(isinstance(st, Categorical))
        self.assertTrue(st.ordered)

        st = hstack([c2, c1])
        self.assertTrue(isinstance(st, Categorical))
        self.assertFalse(st.ordered)

        st = hstack([c1, c2, c1])
        self.assertTrue(isinstance(st, Categorical))
        self.assertTrue(st.ordered)

        st = hstack([c2, c1, c2])
        self.assertTrue(isinstance(st, Categorical))
        self.assertFalse(st.ordered)

    def test_stack_load_ordered(self):
        c1 = Categorical([1, 2, 3], ["a", "b", "c"])
        self.assertTrue(c1.ordered)
        c2 = Categorical([1, 2, 3], ["b", "a", "z"])
        self.assertFalse(c2.ordered)

        ds1 = Dataset({"catcol": c1})
        ds2 = Dataset({"catcol": c2})
        paths = [r"riptable/tests/temp/ds1.sds", r"riptable/tests/temp/ds2.sds"]
        ds1.save(paths[0])
        ds2.save(paths[1])

        ds1 = load_sds(paths[0])
        self.assertTrue(isinstance(ds1.catcol, Categorical))
        self.assertTrue(ds1.catcol.ordered)

        ds2 = load_sds(paths[1])
        self.assertTrue(isinstance(ds2.catcol, Categorical))
        self.assertFalse(ds2.catcol.ordered)

        ds3 = load_sds([paths[1], paths[0]], stack=True)
        self.assertTrue(isinstance(ds3.catcol, Categorical))
        self.assertFalse(ds3.catcol.ordered)

        ds4 = load_sds([paths[0], paths[1]], stack=True)
        self.assertTrue(isinstance(ds4.catcol, Categorical))
        self.assertTrue(ds4.catcol.ordered)

        for p in paths:
            os.remove(p)

    def test_stack_load_sort_display(self):
        c = Categorical([1, 1, 2, 3, 1], ["b", "c", "a"])
        self.assertFalse(c.ordered)
        self.assertFalse(c.grouping.isdisplaysorted)

        d = Categorical([1, 1, 2, 3, 1], ["b", "c", "a"], sort_gb=True)
        self.assertFalse(d.ordered)
        self.assertTrue(d.grouping.isdisplaysorted)

        ds1 = Dataset({"catcol": c})
        ds2 = Dataset({"catcol": d})
        paths = [r"riptable/tests/temp/ds1.sds", r"riptable/tests/temp/ds2.sds"]

        ds1.save(paths[0])
        ds2.save(paths[1])

        ds3 = load_sds(paths, stack=True)
        gb_result = ds3.catcol.count()
        self.assertFalse(issorted(gb_result.catcol))

        ds4 = load_sds([paths[1], paths[0]], stack=True)
        gb_result = ds4.catcol.count()
        self.assertTrue(issorted(gb_result.catcol))

        for p in paths:
            os.remove(p)

    def testapply_reduce(self):
        c = Cat(ones(10))
        # make sure kwargs handled also
        ds = c.apply_reduce(np.std, arange(10), ddof=4)
        # first col, first val
        finalcalc = ds[0][0]
        self.assertTrue(finalcalc > 3.6 and finalcalc < 3.8)

    def testapply_nonreduce(self):
        # even / odd groups
        c = Cat(arange(10) % 2)

        ds = c.apply_nonreduce(np.cumsum, arange(10))

        finalcalc = ds[0]
        goodresult = FA([0, 1, 2, 4, 6, 9, 12, 16, 20, 25])

        self.assertTrue(np.all(finalcalc == goodresult))

        # try with a numba function
        a = arange(10.0)
        a[[3, 4, 7]] = nan

        finalcalc = c.apply_nonreduce(fill_forward, a)[0]
        goodresult = FA([0.0, 1.0, 2.0, 1.0, 2.0, 5.0, 6.0, 5.0, 8.0, 9.0])
        self.assertTrue(np.all(finalcalc == goodresult))

        # try with filtered
        c = Cat([1, 0, 1, 0], ["a"])
        r = c.apply_nonreduce(np.cumsum, FA([10, 1, 20, 2]))
        assert_array_equal(r[0], FA([10, 1, 30, 3]))

    def test_ema(self):
        arrsize = 100
        numrows = 20
        ds = Dataset({"time": arange(arrsize * 1.0)})
        ds.data = np.random.randint(numrows, size=arrsize)
        ds.data2 = np.random.randint(numrows, size=arrsize)
        symbols = ["ZYGO", "YHOO", "FB", "GOOG", "IBM"]
        ds.symbol2 = Cat(
            1 + ds.data,
            [
                "A",
                "B",
                "C",
                "D",
                "E",
                "F",
                "G",
                "H",
                "I",
                "J",
                "K",
                "L",
                "M",
                "N",
                "O",
                "P",
                "Q",
                "R",
                "S",
                "T",
            ],
        )
        ds.symbol = Cat(1 + arange(arrsize) % len(symbols), symbols)
        result = ds.data.ema_decay(ds.time, 2.2)

        result = ds.data.ema_decay(ds.time, 2.2, filter=ds.symbol == "YHOO")

        gb = ds.gb("symbol")
        prev = gb.grouping.iprevkey
        next = gb.grouping.inextkey

    def test_bincount(self):
        c = np.repeat(arange(1000), 3000)
        c1 = Cat(c)
        nCountGroup = rc.BinCount(c1._fa, c1.unique_count + 1)
        y = groupbypack(c1._fa, None, c1.unique_count + 1)
        # all of them should be 30000 except invalid bin
        x = (nCountGroup == 3000).sum()
        self.assertTrue(x == 1000)
        self.assertTrue(np.all(nCountGroup == y["nCountGroup"]))
        z = bincount(c)
        self.assertTrue(np.all(nCountGroup[1:] == z))

        c = np.random.randint(0, 1000, 1_000_000)
        x = groupbyhash(c)
        ncountgroup = rc.BinCount(x["iKey"], x["unique_count"] + 1)
        y = groupbypack(x["iKey"], ncountgroup)
        self.assertTrue(np.all(ncountgroup == y["nCountGroup"]))

        c1 = Cat(c)
        pack = groupbypack(x["iKey"], None, c1.unique_count + 1)
        self.assertTrue(np.all(ncountgroup == pack["nCountGroup"]))
        pack2 = groupbypack(x["iKey"], ncountgroup, None)
        self.assertTrue(np.all(pack2["iFirstGroup"] == pack["iFirstGroup"]))

        ## every 1 in 1000 is unique
        # c1=Cat(c)
        # nCat = c1.unique_count + 1
        # dCat = c1._fa
        # nCountGroup = rc.BinCount(c1._fa, c1.unique_count +1 )
        # x= rc.GroupFromBinCount(c1._fa, nCountGroup)

        ## highly unique count
        # c=np.random.randint(0, 300_000, 1_000_000)
        # c1=Cat(c)
        # nCat = c1.unique_count + 1
        # dCat = c1._fa

        ## Force int64
        # c=np.repeat(arange(1000),3000)
        # c1=Cat(c)
        # nCat = c1.unique_count + 1
        # dCat = c1._fa
        # nCountGroup = rc.BinCount(c1._fa, c1.unique_count +1 )

    def test_take_groups(self):
        ## Case 1: Basic operation.
        # Create a grouping from some data.
        key_data1 = rt.FA([1, 2, 2, 3, 3, 3, 4, 4, 4, 4, 5, 5, 5, 5, 5, 6, 6, 6, 6, 6, 6])
        g1 = rt.Grouping(key_data1)

        # Create another data array the same length as the key for the Grouping.
        data1 = rt.arange(len(key_data1))

        # Extract elements from the data array where they correspond to even-numbered groups.
        result1 = Grouping.take_groups(data1, rt.FA([2, 4, 6]), g1.ncountgroup, g1.ifirstgroup)

        assert_array_equal(rt.FA([1, 2, 6, 7, 8, 9, 15, 16, 17, 18, 19, 20]), result1)

    def test_extract_groups(self):
        ## Case 1: Basic operation.
        # Create a grouping from some data.
        key_data1 = rt.FA([1, 2, 2, 3, 3, 3, 4, 4, 4, 4, 5, 5, 5, 5, 5, 6, 6, 6, 6, 6, 6])
        g1 = rt.Grouping(key_data1)

        # Create another data array the same length as the key for the Grouping.
        data1 = rt.arange(len(key_data1))

        # Create a mask which selects the even-numbered groups.
        group_mask1 = rt.arange(len(g1.ncountgroup)) % 2 == 0

        # Extract elements from the data array where they correspond to even-numbered groups.
        result1 = Grouping.extract_groups(group_mask1, data1, g1.ncountgroup, g1.ifirstgroup)

        assert_array_equal(rt.FA([1, 2, 6, 7, 8, 9, 15, 16, 17, 18, 19, 20]), result1)

    def test_extract_groups_all_groups_off(self):
        """Test for Grouping.extract_groups() when given a condition mask with all values set to False."""

        # Create a grouping from some data.
        key_data1 = rt.FA([1, 2, 2, 3, 3, 3, 4, 4, 4, 4, 5, 5, 5, 5, 5, 6, 6, 6, 6, 6, 6])
        g1 = rt.Grouping(key_data1)

        # Create another data array the same length as the key for the Grouping.
        data1 = rt.arange(len(key_data1))

        # Create a condition mask with all values set to False.
        group_mask1 = rt.zeros(len(g1.ncountgroup), dtype=bool)

        # Extract elements from the data array with all groups masked out -- i.e. we're trying
        # to select data from none of the groups.
        result1 = Grouping.extract_groups(group_mask1, data1, g1.ncountgroup, g1.ifirstgroup)

        assert_array_equal(rt.FA([]), result1)


def test_merge_cats_stringcat_with_empty():
    cat_lens = [
        12,
        7,
    ]  # Arbitrary; specify the lengths of our test Cats here so we can re-use the lengths below for consistency.

    indices = np.hstack([np.full(cat_lens[0], 1), np.full(cat_lens[1], 2)])
    listcats = [rt.FA([b"2019/12/21", b"", b"2019/12/21"], dtype="|S32")]
    idx_cutoffs = np.cumsum(cat_lens)
    uniques_cutoffs = [np.array([1, 3], dtype=np.int64)]
    assert len(listcats) == len(uniques_cutoffs)

    fixed_indices, stacked_uniques = rt.merge_cats(
        indices, listcats, idx_cutoffs=idx_cutoffs, unique_cutoffs=uniques_cutoffs
    )

    assert len(stacked_uniques[0]) == 2


if __name__ == "__main__":
    tester = unittest.main()



import unittest

import numpy as np
import pytest
from numpy.testing import assert_array_equal

import riptable as rt
from riptable import FastArray

rand_nums = np.random.choice(30, 100)
numeric_types = "fd" + "bB" + "hH" + "iIlL" + "qQpP"  # + '?'


@pytest.mark.parametrize("dt_char", list(numeric_types))
@pytest.mark.parametrize("lex", [False, True])
def test_numeric_accuracy(dt_char, lex):
    target_dtype = np.dtype(dt_char)

    # N.B. It's important to use .view() instead of .astype() here so the case
    #      for np.bool_ is tested thoroughly; .astype(bool) will convert the
    #      values to 0/1, but .view(bool) will not -- so it will check that
    #      lower-level logic does proper C-style bool conversion (zero=false, non-zero=true).
    nums = rand_nums.view(target_dtype)
    np_un, np_idx, np_inv, np_cnt = np.unique(nums, return_index=True, return_inverse=True, return_counts=True)
    fa_un, fa_idx, fa_inv, fa_cnt = rt.unique(nums, return_index=True, return_inverse=True, return_counts=True, lex=lex)

    assert_array_equal(
        np_un,
        fa_un,
        err_msg=f"Results did not match for unique items for dtype {target_dtype.name}.",
    )
    assert_array_equal(
        np_idx,
        fa_idx,
        err_msg=f"Results did not match for unique fancy index for dtype {target_dtype.name}.",
    )
    assert_array_equal(
        np_inv,
        fa_inv,
        err_msg=f"Results did not match for inverse fancy index for dtype {target_dtype.name}.",
    )
    assert_array_equal(
        np_cnt,
        fa_cnt,
        err_msg=f"Results did not match for unique counts for dtype {target_dtype.name}.",
    )


@pytest.mark.parametrize("dt_char", list("US"))
@pytest.mark.parametrize("lex", [False, True])
def test_string_accuracy(dt_char, lex):
    target_dtype = np.dtype(dt_char)

    rand_strings = (
        np.random.choice(["test_string" + str(i) for i in range(30)], 100)
        .astype(target_dtype, copy=False)
        .view(FastArray)
    )
    np_un, np_idx, np_inv, np_cnt = np.unique(rand_strings, return_index=True, return_inverse=True, return_counts=True)
    fa_un, fa_idx, fa_inv, fa_cnt = rt.unique(
        rand_strings,
        return_index=True,
        return_inverse=True,
        return_counts=True,
        lex=lex,
    )

    assert_array_equal(
        np_un,
        fa_un,
        err_msg=f"Results did not match for unique items for dtype {target_dtype.name}.",
    )
    assert_array_equal(
        np_idx,
        fa_idx,
        err_msg=f"Results did not match for unique fancy index for dtype {target_dtype.name}.",
    )
    assert_array_equal(
        np_inv,
        fa_inv,
        err_msg=f"Results did not match for inverse fancy index for dtype {target_dtype.name}.",
    )
    assert_array_equal(
        np_cnt,
        fa_cnt,
        err_msg=f"Results did not match for unique counts for dtype {target_dtype.name}.",
    )


# TODO: Once the underlying logic in Grouping and/or rt.unique() has been fixed, add '?'
#       back to the 'numeric_types' variable above, then this test can be removed
#       (since bools will be included in the test_numeric_accuracy test).
@pytest.mark.xfail(reason="RIP-371: Grouping returns incorrect results for bool array")
@pytest.mark.parametrize("dt_char", list("?"))
@pytest.mark.parametrize("lex", [False, True])
def test_bool_accuracy(dt_char, lex):
    target_dtype = np.dtype(dt_char)

    # N.B. It's important to use .view() instead of .astype() here so the case
    #      for np.bool is tested thoroughly; .astype(bool) will convert the
    #      values to 0/1, but .view(bool) will not -- so it will check that
    #      lower-level logic does proper C-style bool conversion (zero=false, non-zero=true).
    nums = rand_nums.view(target_dtype)
    np_un, np_idx, np_inv, np_cnt = np.unique(nums, return_index=True, return_inverse=True, return_counts=True)
    fa_un, fa_idx, fa_inv, fa_cnt = rt.unique(nums, return_index=True, return_inverse=True, return_counts=True, lex=lex)

    assert_array_equal(
        np_un,
        fa_un,
        err_msg=f"Results did not match for unique items for dtype {target_dtype.name}.",
    )
    assert_array_equal(
        np_idx,
        fa_idx,
        err_msg=f"Results did not match for unique fancy index for dtype {target_dtype.name}.",
    )
    assert_array_equal(
        np_inv,
        fa_inv,
        err_msg=f"Results did not match for inverse fancy index for dtype {target_dtype.name}.",
    )
    assert_array_equal(
        np_cnt,
        fa_cnt,
        err_msg=f"Results did not match for unique counts for dtype {target_dtype.name}.",
    )


if __name__ == "__main__":
    tester = unittest.main()



"""
Functions for generating random data for use in benchmarking.
"""
__all__ = ["rand_array", "rand_dataset", "rand_fancyindex", "rand_keyarray"]

from typing import List, Optional, Union

import numpy as np

from ..rt_categorical import Categorical
from ..rt_dataset import Dataset
from ..rt_fastarray import FastArray
from ..rt_numpy import arange, putmask


def check_params(dtype, invalid_ratio):
    if not isinstance(dtype, np.dtype):
        raise TypeError(
            f"The argument provided for the `dtype` parameter has the type '{type(dtype)}' but only instances of numpy.dtype are allowed."
        )
    elif invalid_ratio is not None:
        if 0.0 <= invalid_ratio <= 1.0:
            pass
        else:
            raise ValueError(f"Invalid value specified for `invalid_ratio`: {invalid_ratio}")


def rand_fancyindex(
    rng: np.random.Generator,
    index_length: int,
    dtype: np.dtype,
    source_arr_len: int,
    invalid_ratio: Optional[float] = None,
) -> np.ndarray:
    """Create a random fancy index with the specified length and dtype."""
    check_params(dtype, invalid_ratio)
    if dtype.kind not in "iu":  # TODO: Also support floats, since mbget allows that
        raise ValueError(f"Only integer dtypes are currently supported by this method. dtype={dtype.name}")

    # Generate the fancy index from the uniform integer distribution.
    fancyindex = FastArray(rng.integers(0, source_arr_len, size=index_length, dtype=dtype))

    # If the fancy index should have some invalids/NA values, add those in now.
    if invalid_ratio is not None and invalid_ratio > 0.0:
        # TODO: Also add in some out-of-bounds accesses (and not just invalid/NA values) here?
        invalid_outcomes = FastArray(rng.random(size=index_length))
        putmask(fancyindex, invalid_outcomes < invalid_ratio, fancyindex.inv)

    return fancyindex


def rand_array(
    rng: np.random.Generator, length: int, dtype: np.dtype, invalid_ratio: Optional[float] = None
) -> np.ndarray:
    # TODO: Implement a flag that controls whether invalid values are included in the array? Or (instead) an invalid_ratio parameter like our other functions?
    check_params(dtype, invalid_ratio)

    if dtype.kind in "iu":
        info = np.iinfo(dtype)
        arr = FastArray(rng.integers(info.min, info.max, size=length, dtype=dtype))

    elif dtype.kind == "f":
        # PERF: Use an FMA function here if we ever implement one
        arr = (FastArray(rng.random(size=length, dtype=dtype)) * 1e10) - 0.5e10

    elif dtype.kind == "S":
        # Generate integers in the upper ASCII range, then use a view to expose those
        # values as fixed-length ASCII strings.
        # TODO: Support other character ranges (lower-range ASCII 0-127, full ASCII 0-255, lowercase+uppercase+digits).
        arr = FastArray(rng.integers(65, 90, size=length * dtype.itemsize, dtype=np.int8, endpoint=True).view(dtype))

    elif dtype.kind == "U":
        # Generate integers in the upper ASCII range.
        # TODO: Support other character ranges (lower-range ASCII 0-127, full ASCII 0-255, lowercase+uppercase+digits, Unicode chars >255).
        arr = FastArray(
            rng.integers(65, 90, size=length * (dtype.itemsize // 4), dtype=np.int32, endpoint=True).view(dtype)
        )

    else:
        # TODO: Handle other dtypes
        raise NotImplementedError(f"The dtype {dtype} is not yet supported by this function.")

    # If the fancy index should have some invalids/NA values, add those in now.
    if invalid_ratio is not None and invalid_ratio > 0.0:
        # TODO: Also add in some out-of-bounds accesses (and not just invalid/NA values) here?
        invalid_outcomes = FastArray(rng.random(size=length))
        putmask(arr, invalid_outcomes < invalid_ratio, arr.inv)

    return arr


def rand_keyarray(
    rng: np.random.Generator,
    length: int,
    dtype: np.dtype,
    unique_count: int,
    invalid_ratio: Optional[float] = None,
) -> np.ndarray:
    """
    Generate a random array representing an array to be used as an input to a set-like
    operation (ismember, groupbyhash, etc.)

    Parameters
    ----------
    rng : np.random.Generator
    length : int
    dtype : np.dtype
    unique_count : int
    invalid_ratio : float, optional

    Returns
    -------
    np.ndarray

    Notes
    -----
    TODO: additional, advanced options:
      * different distributions (e.g. rng.zipf()) for key multiplicity (i.e. when creating the fancy index we use to
        fetch from the unique values array).
      * key clustering (i.e. are keys already more or less occurring near each other, or are they all spread out?)
      * key dispersion -- if the unique keys are created like arange(1, 100), does grouping/merge go any
        faster than if the keys are created like arange(1, 1000, 10) or arange(1, 1_000_000_000, 10_000_000_000)?
        Make sure to control for dtype to ensure that's the same for all cases tested here.
        This could be used to diagnose issues with hashing/grouping implementations.
      * iKey sortedness
    """
    if not isinstance(dtype, np.dtype):
        raise TypeError(
            f"The argument provided for the `dtype` parameter has the type '{type(dtype)}' but only instances of numpy.dtype are allowed."
        )

    # Generate array of unique values.
    if dtype.kind in "iu":
        # TODO: Support non-consecutive values here
        unique_values = arange(0, unique_count, dtype=dtype)

    elif dtype.kind == "S":
        # Generate integers in the upper ASCII range, then use a view to expose those
        # values as fixed-length ASCII strings.
        # TODO: Support other character ranges (lower-range ASCII 0-127, full ASCII 0-255, lowercase+uppercase+digits).
        # TODO: Use uniques() or similar to make sure all of the values generated here are actually unique.
        unique_values = rng.integers(65, 90, size=unique_count * dtype.itemsize, dtype=np.int8, endpoint=True).view(
            dtype
        )

    elif dtype.kind == "U":
        # Generate integers in the upper ASCII range.
        # TODO: Support other character ranges (lower-range ASCII 0-127, full ASCII 0-255, lowercase+uppercase+digits, Unicode chars >255).
        # TODO: Use uniques() or similar to make sure all of the values generated here are actually unique.
        unique_values = rng.integers(
            65,
            90,
            size=unique_count * (dtype.itemsize // 4),
            dtype=np.int32,
            endpoint=True,
        ).view(dtype)

    else:
        # TODO: Support other dtypes, e.g. floats
        raise NotImplementedError(f"dtype {dtype} not yet supported by this function.")

    # Generate a fancy index into the array of unique values.
    fancyindex = rand_fancyindex(
        rng,
        length,
        np.min_scalar_type(length),
        unique_count,
        invalid_ratio=invalid_ratio,
    )

    # Use the fancy index to select elements from the array of unique values, then return the result.
    return unique_values[fancyindex]


def rand_multikeyarray(rng: np.random.Generator, length: int, dtypes: List[np.dtype]) -> List[np.ndarray]:
    # TODO: Implement a function similar to rand_keyarray but which produces a multikey (a tuple or list of arrays)
    raise NotImplementedError()


def rand_dataset(
    rowcount: int,
    rng: np.random.Generator,
    # Can the next few params be combined into some "KeySpec" NamedTuple to help simplify passing them?
    # TODO: Allow the key name(s) to be specified?
    unique_key_count: Union[int, List[int]],
    key_dtype: Optional[Union[np.dtype, List[np.dtype]]] = None,
    make_cat: Union[bool, List[bool]] = True,
    xtra_col_dtypes: Optional[List[np.dtype]] = None,
) -> Dataset:
    """
    Create a random Dataset with the specified shape and key(s).

    Parameters
    ----------
    rowcount
    rng
    unique_key_count
    key_dtype
        When None, defaults to np.int64.
    make_cat
    xtra_col_dtypes

    Returns
    -------
    Dataset
    """
    if key_dtype is None:
        key_dtype = np.dtype(np.int64)
    elif isinstance(key_dtype, list):
        for ty in key_dtype:
            if not isinstance(ty, np.dtype):
                raise TypeError(
                    f"The list provided for the `key_dtype` parameter contains an instance of type '{type(ty)}' but only instances of numpy.dtype are allowed."
                )

    elif not isinstance(key_dtype, np.dtype):
        raise TypeError(
            f"The argument provided for the `key_dtype` parameter has the type '{type(key_dtype)}' but only instances of numpy.dtype are allowed."
        )

    ds = Dataset()

    # Create the key column.
    # TODO: Offer an option to use other key distributions like rng.zipf(). Or allow the distribution to be specified per column so we can e.g. have one uniform column and one zipf column, then combine them into a multikey.
    keydata = (
        rand_keyarray(rng, rowcount, key_dtype, unique_key_count)
        if isinstance(key_dtype, np.dtype)
        else rand_multikeyarray(rng, rowcount, key_dtype)
    )

    # Create a Categorical from this data if specified.
    if make_cat:
        keydata = Categorical(keydata)

    # TODO: If keydata is a list here, we need to assign the columns into the Dataset individually.
    ds["key"] = keydata

    # Add a column with the row index (just so there's at least one additional column besides the key).
    ds["rowidx"] = arange(rowcount)

    # Add additional columns with random data if requested.
    if xtra_col_dtypes:
        raise NotImplementedError()

    return ds



__all__ = ["fill_forward", "fill_backward"]

# ---------------------------------------------------------------------
# This file is for numba routines that work on numpy arrays
# It is different from CPP routines in the riptide_cpp module
# Others are encouraged to add
#
import warnings
from typing import Optional, Union

import numba as nb
import numpy as np

from .config import get_global_settings
from .rt_enum import INVALID_DICT
from .rt_fastarray import FastArray
from .rt_numpy import empty_like


# -----------------------------------------------------
@nb.njit(parallel=False, cache=get_global_settings().enable_numba_cache, nogil=True)
def fill_forward_float(arr, fill_val, limit):
    lastgood = fill_val
    if limit <= 0:
        for idx in range(arr.shape[0]):
            if np.isnan(arr[idx]):
                arr[idx] = lastgood
            else:
                lastgood = arr[idx]
    else:
        counter = limit
        for idx in range(arr.shape[0]):
            if np.isnan(arr[idx]):
                # leave the value if counter <= 0
                if counter > 0:
                    arr[idx] = lastgood
                    counter -= 1
            else:
                # reset counter
                counter = limit
                lastgood = arr[idx]


# -----------------------------------------------------
@nb.njit(parallel=False, cache=get_global_settings().enable_numba_cache, nogil=True)
def fill_forward_int(arr, inv, fill_val, limit):
    lastgood = fill_val
    if limit <= 0:
        for idx in range(arr.shape[0]):
            if arr[idx] == inv:
                arr[idx] = lastgood
            else:
                lastgood = arr[idx]
    else:
        counter = limit
        for idx in range(arr.shape[0]):
            if arr[idx] == inv:
                # leave the value if counter <= 0
                if counter > 0:
                    arr[idx] = lastgood
                    counter -= 1
            else:
                # reset counter
                counter = limit
                lastgood = arr[idx]


# -----------------------------------------------------
@nb.njit(parallel=False, cache=get_global_settings().enable_numba_cache, nogil=True)
def fill_backward_float(arr, fill_val, limit):
    lastgood = fill_val
    if limit <= 0:
        for idx in range(arr.shape[0] - 1, -1, -1):
            if np.isnan(arr[idx]):
                arr[idx] = lastgood
            else:
                lastgood = arr[idx]
    else:
        counter = limit
        for idx in range(arr.shape[0] - 1, -1, -1):
            if np.isnan(arr[idx]):
                # leave the value if counter <= 0
                if counter > 0:
                    arr[idx] = lastgood
                    counter -= 1
            else:
                # reset counter
                counter = limit
                lastgood = arr[idx]


# -----------------------------------------------------
@nb.njit(parallel=False, cache=get_global_settings().enable_numba_cache, nogil=True)
def fill_backward_int(arr, inv, fill_val, limit):
    lastgood = fill_val
    if limit <= 0:
        for idx in range(arr.shape[0] - 1, -1, -1):
            if arr[idx] == inv:
                arr[idx] = lastgood
            else:
                lastgood = arr[idx]
    else:
        counter = limit
        for idx in range(arr.shape[0] - 1, -1, -1):
            if arr[idx] == inv:
                # leave the value if counter <= 0
                if counter > 0:
                    arr[idx] = lastgood
                    counter -= 1
            else:
                # reset counter
                counter = limit
                lastgood = arr[idx]


# -----------------------------------------------------
def _check_fill_values(arr, fill_val, inplace: bool, limit: int):
    if arr.dtype.num > 13 or arr.dtype.num == 0:
        # fill string, boolean, other?
        raise TypeError(f"Filling for type {type(arr)} is currently not supported.")

    if limit is None:
        limit = 0

    limit = np.int64(limit)

    if limit < 0:
        raise TypeError(f"The limit kwarg cannot be less than 0.")

    if not inplace:
        arr = arr.copy()

    dtype = arr.dtype
    inv = INVALID_DICT[dtype.num]
    inv = np.array([inv], dtype=dtype)[0]

    if fill_val is None:
        # optionally could raise error
        fill_val = inv

    # force into np scalar with dtype
    # TODO there is a better way to do this
    fill_val = np.array([fill_val], dtype=dtype)[0]

    return arr, fill_val, inv, dtype, limit


# -----------------------------------------------------
def fill_forward(arr: np.ndarray, fill_val=None, inplace: bool = False, limit: int = 0):
    """
    Replace NaN and invalid array values by propagating the last encountered valid value
    forward.

    Note that this method can be called either as a `FastArray` method
    (:meth:`riptable.rt_fastarraynumba.fill_forward`) or a function
    (:meth:`riptable.fill_forward()`) that takes an array or `FastArray` as input. The
    function returns either an array or a `FastArray`, depending on the original input.

    Parameters
    ----------
    fill_val : scalar, default None
        The value to use where there is no valid value to propagate forward. If
        `fill_val` is not specified, NaN and invalid values aren't replaced where there
        is no valid value to propagate forward.
    inplace : bool, default False
        If False, return a copy of the array. If True, modify original data. This will
        modify any other views on this object.
    limit : int, default 0
        The maximium number of consecutive NaN or invalid values to fill. If there is a
        gap with more than this number of consecutive NaN or invalid values, the gap
        will be only partially filled. If no `limit` is specified, all consecutive NaN
        and invalid values are replaced.

    Returns
    -------
    `FastArray`
        The `FastArray` will be the same size and have the same dtype as the original
        input.

    See Also
    --------
    riptable.rt_fastarraynumba.fill_backward : Replace NaN and invalid values with the
        next valid value.
    riptable.rt_fastarraynumba.fill_forward : Replace NaN and invalid values with the
        last valid value.
    riptable.fill_forward : Replace NaN and invalid values with the last valid value.
    FastArray.fillna : Replace NaN and invalid values with a specified value or nearby
        data.
    FastArray.replacena : Replace NaN and invalid values with a specified value.
    Dataset.fillna : Replace NaN and invalid values with a specified value or nearby
        data.
    Categorical.fill_forward : Replace NaN and invalid values with the last valid
        group value.
    GroupBy.fill_forward : Replace NaN and invalid values with the last valid group
        value.

    Examples
    --------
    Use a `fill_val` to replace values where there's no valid value to propagate
    forward:

    >>> a = rt.FastArray([rt.nan, 1.0, rt.nan, rt.nan, rt.nan, 5.0])
    >>> a.fill_forward(fill_val = 0)
    FastArray([0., 1., 1., 1., 1., 5.])

    Using :meth:`riptable.fill_forward`:

    >>> a = rt.FastArray([0.0, rt.nan, rt.nan, rt.nan, 4.0, rt.nan])
    >>> rt.fill_forward(a, fill_val = 0)
    FastArray([0., 0., 0., 0., 4., 4.])

    Replace only the first NaN or invalid value in any consecutive series of NaN or
    invalid values:

    >>> a.fill_forward(limit = 1)
    FastArray([nan,  1.,  1., nan, nan,  5.])
    """
    # TODO: handle axis
    arr, fill_val, inv, dtype, limit = _check_fill_values(arr, fill_val, inplace, limit)

    if dtype.num <= 10:
        # fill integers or boolean
        fill_forward_int(arr, inv, fill_val, limit)
    else:
        # fill float
        fill_forward_float(arr, fill_val, limit)

    return arr


# -----------------------------------------------------
def fill_backward(arr: np.ndarray, fill_val=None, inplace: bool = False, limit: int = 0):
    """
    Replace NaN and invalid array values by propagating the next encountered valid value
    backward.

    Note that this method can be called either as a `FastArray` method
    (:meth:`riptable.rt_fastarraynumba.fill_backward`) or a function
    (:meth:`riptable.fill_backward`) that takes an array or `FastArray` as input. The
    function returns either an array or a `FastArray`, depending on the original input.

    Parameters
    ----------
    fill_val : scalar, default None
        The value to use where there is no valid value to propagate backward. If
        `fill_val` is not specified, NaN and invalid values aren't replaced where there
        is no valid value to propagate backward.
    inplace : bool, default False
        If False, return a copy of the array. If True, modify original data. This will
        modify any other views on this object.
    limit : int, default 0
        The maximium number of consecutive NaN or invalid values to fill. If there is a
        gap with more than this number of consecutive NaN or invalid values, the gap
        will be only partially filled. If no `limit` is specified, all consecutive NaN
        and invalid values are replaced.

    Returns
    -------
    `FastArray`
        The `FastArray` will be the same size and have the same dtype as the original
        input.

    See Also
    --------
    riptable.rt_fastarraynumba.fill_forward : Replace NaN and invalid values with the
        last valid value.
    riptable.rt_fastarraynumba.fill_backward : Replace NaN and invalid values with the
        next valid value.
    riptable.fill_backward : Replace NaN and invalid values with the next valid value.
    FastArray.fillna : Replace NaN and invalid values with a specified value or nearby
        data.
    FastArray.replacena : Replace NaN and invalid values with a specified value.
    Dataset.fillna : Replace NaN and invalid values with a specified value or nearby
        data.
    Categorical.fill_backward : Replace NaN and invalid values with the next valid
        group value.
    GroupBy.fill_backward : Replace NaN and invalid values with the next valid group
        value.

    Examples
    --------
    Use a `fill_val` to replace values where there's no valid value to propagate
    backward:

    >>> a = rt.FastArray([0.0, rt.nan, rt.nan, rt.nan, 4.0, rt.nan])
    >>> a.fill_backward(fill_val = 0)
    FastArray([0., 4., 4., 4., 4., 0.])

    Using :meth:`riptable.fill_backward`:

    >>> a = rt.FastArray([0.0, rt.nan, rt.nan, rt.nan, 4.0, rt.nan])
    >>> rt.fill_backward(a, fill_val = 0)
    FastArray([0., 4., 4., 4., 4., 0.])

    Replace only the first NaN or invalid value in any consecutive series of NaN or
    invalid values:

    >>> a.fill_backward(limit = 1)
    FastArray([ 0., nan, nan,  4.,  4., nan])
    """
    # TODO: handle axis
    arr, fill_val, inv, dtype, limit = _check_fill_values(arr, fill_val, inplace, limit)

    if dtype.num <= 10:
        # fill integers or boolean
        fill_backward_int(arr, inv, fill_val, limit)
    else:
        # fill float
        fill_backward_float(arr, fill_val, limit)

    return arr


# -----------------------------------------------------
@nb.njit(parallel=False, cache=get_global_settings().enable_numba_cache, nogil=True)
def nb_cummin_int(arr: np.ndarray, ret: np.ndarray, inv, skipna):
    if skipna:
        for j in range(len(arr)):
            running_min = arr[j]
            ret[j] = running_min
            if running_min != inv:
                break

        for i in range(j, len(arr)):
            val = arr[i]
            if val != inv and val < running_min:
                running_min = val
            ret[i] = running_min
    else:
        running_min = arr[0]
        for i in range(len(arr)):
            val = arr[i]
            if val == inv or val < running_min:
                running_min = val
            ret[i] = running_min


# -----------------------------------------------------
@nb.njit(parallel=False, cache=get_global_settings().enable_numba_cache, nogil=True)
def nb_cummin_float(arr: np.ndarray, ret: np.ndarray, skipna):
    if skipna:
        for j in range(len(arr)):
            running_min = arr[j]
            ret[j] = running_min
            if running_min == running_min:
                break

        for i in range(j, len(arr)):
            val = arr[i]
            if val == val and val < running_min:
                running_min = val
            ret[i] = running_min
    else:
        running_min = arr[0]
        for i in range(len(arr)):
            val = arr[i]
            if val != val or val < running_min:
                running_min = val
            ret[i] = running_min


# -----------------------------------------------------
@nb.njit(parallel=False, cache=get_global_settings().enable_numba_cache, nogil=True)
def nb_cummax_int(arr: np.ndarray, ret: np.ndarray, inv, skipna):
    if skipna:
        for j in range(len(arr)):
            running_max = arr[j]
            ret[j] = running_max
            if running_max != inv:
                break

        for i in range(j, len(arr)):
            val = arr[i]
            if val != inv and val > running_max:
                running_max = val
            ret[i] = running_max
    else:
        running_max = arr[0]
        for i in range(len(arr)):
            val = arr[i]
            if val == inv or val > running_max:
                running_max = val
            ret[i] = running_max


# -----------------------------------------------------
@nb.njit(parallel=False, cache=get_global_settings().enable_numba_cache, nogil=True)
def nb_cummax_float(arr: np.ndarray, ret: np.ndarray, skipna):
    if skipna:
        for j in range(len(arr)):
            running_max = arr[j]
            ret[j] = running_max
            if running_max == running_max:
                break

        for i in range(j, len(arr)):
            val = arr[i]
            if val == val and val > running_max:
                running_max = val
            ret[i] = running_max
    else:
        running_max = arr[0]
        for i in range(len(arr)):
            val = arr[i]
            if val != val or val > running_max:
                running_max = val
            ret[i] = running_max


def cummax(arr: np.ndarray, skipna=True):
    """
    Return the running maximum over an array.

    Parameters
    ----------
    skipna : boolean, default True
        Exclude nan/invalid values.

    By default, nan values are ignored.

    Examples
    --------
    a=rt.FA([1,2,3,4,2,3,5,5,6,2,7])
    >>> a.cummax()
    FastArray([1, 2, 3, 4, 4, 4, 5, 5, 6, 6, 7])

    >>> a[a >= a.cummax()]
    FastArray([1, 2, 3, 4, 5, 5, 6, 7])

    >>> a[1]=rt.nan
    >>> a.cummax()
    FastArray([1, 1, 3, 4, 4, 4, 5, 5, 6, 6, 7])

    See Also
    --------
    cummin, cumprod, cumsum
    """
    ret = empty_like(arr)
    if len(arr) > 0:
        dtype = arr.dtype
        inv = INVALID_DICT[dtype.num]

        if dtype.num <= 10:
            nb_cummax_int(arr, ret, inv, skipna)
        elif dtype.num <= 13:
            nb_cummax_float(arr, ret, skipna)
        else:
            raise TypeError("cummax only handles integers or floats.")

    return ret


def cummin(arr: np.ndarray, skipna=True):
    """
    Return the running minimum over an array.

    Parameters
    ----------
    skipna : boolean, default True
        Exclude nan/invalid values.

    By default, nan values are ignored.

    Examples
    --------
    >>> a=FA([1,2,3,-4,2,3,5,5,-6,2,7])
    >>> a.cummin()
    FastArray([ 1,  1,  1, -4, -4, -4, -4, -4, -6, -6, -6])

    See Also
    --------
    cummax, cumprod, cumsum
    """
    ret = empty_like(arr)
    if len(arr) > 0:
        dtype = arr.dtype
        inv = INVALID_DICT[dtype.num]

        if dtype.num <= 10:
            nb_cummin_int(arr, ret, inv, skipna)
        elif dtype.num <= 13:
            nb_cummin_float(arr, ret, skipna)
        else:
            raise TypeError("cummin only handles integers or floats.")

    return ret


# -----------------------------------------------------
@nb.njit(parallel=False, cache=get_global_settings().enable_numba_cache, nogil=True)
def nb_ema_decay_with_filter_and_reset(arr, dest, time, decayRate, filter, resetmask):
    lastEma = 0
    lastTime = 0
    for i in range(len(arr)):
        value = 0

        # NOTE: fill in last value
        if filter[i] != 0:
            value = arr[i]

            if resetmask[i]:
                lastEma = 0
                lastTime = 0

        lastEma = value + lastEma * np.exp(-decayRate * (time[i] - lastTime))
        lastTime = time[i]
        dest[i] = lastEma


# -----------------------------------------------------
@nb.njit(parallel=False, cache=get_global_settings().enable_numba_cache, nogil=True)
def nb_ema_decay_with_filter(arr, dest, time, decayRate, filter):
    lastEma = 0
    lastTime = 0
    for i in range(len(arr)):
        value = 0

        # NOTE: fill in last value
        if filter[i] != 0:
            value = arr[i]

        lastEma = value + lastEma * np.exp(-decayRate * (time[i] - lastTime))
        lastTime = time[i]
        dest[i] = lastEma


# -----------------------------------------------------
@nb.njit(parallel=False, cache=get_global_settings().enable_numba_cache, nogil=True)
def nb_ema_decay(arr, dest, time, decayRate):
    lastEma = 0.0
    lastTime = 0
    for i in range(len(arr)):
        value = arr[i]

        # nan check (assumes float nans and not invalids)
        if value != value:
            value = 0

        lastEma = value + lastEma * np.exp(-decayRate * (time[i] - lastTime))
        lastTime = time[i]
        dest[i] = lastEma


# -----------------------------------------------------
def ema_decay(
    arr: np.ndarray,
    time: np.ndarray,
    decay_rate: float,
    filter: Optional[np.ndarray] = None,
    reset: Optional[np.ndarray] = None,
    dtype=np.float32,
):
    """
    Calculate the EMA using a fixed decay rate.

    Parameters
    ----------
    arr : array-like
        An array whose value is decayed over time.
    time : array-like
        A time array of equal length to `arr`. Often an int64 array.
    decay_rate : float
        A scalar value float such as 2.27.
    dtype : string or data-type
        The dtype to force for the output array. Defaults to ``np.float32``, can change to ``np.float64`` for more accuracy.

    Other Parameters
    ----------------
    filter : array-like, optional
        A boolean mask. If supplied, must be the same length as `arr`.
    reset : array-like, optional
        A boolean mask. Only valid if `filter` is set.

    Returns
    -------
    FastArray
        A float array of EMA values. The dtype of this array is specified by the `dtype` parameter.

    Examples
    --------
    `ema_decay` with a decay factor of 0 means "no decay", which means the time component
    is effectively ignored; in this case, `ema_decay` gives the same result as a cumulative-sum.

    >>> data = rt.ones(10)
    >>> times = rt.FastArray([0, 1, 1, 3, 4, 5, 5.5, 10.5, 10.55, 11])
    >>> rt.FastArray.ema_decay(data, times, 0)
    FastArray([ 1.,  2.,  3.,  4.,  5.,  6.,  7.,  8.,  9., 10.],
            dtype=float32)

    Simple use case with a 50% decay rate per time-unit.

    >>> rt.FastArray.ema_decay(data, times, 0.5)
    FastArray([1.       , 1.6065307, 2.6065307, 1.958889 , 2.1881263,
            2.3271656, 2.8123984, 1.2308557, 2.2004657, 2.7571077],
            dtype=float32)

    Specify the `dtype` argument explicitly to force the type of the output array
    for additional precision when needed.

    >>> rt.FastArray.ema_decay(data, times, 0.5, dtype=np.float64)
    FastArray([1.        , 1.60653066, 2.60653066, 1.95888904, 2.18812626,
            2.32716567, 2.81239844, 1.23085572, 2.20046579, 2.75710762])

    Provide a filter (boolean mask) to mask out elements which should be skipped over
    during the EMA calculation. Whenever an element has been masked out / skipped,
    the corresponding element of the output array will be the same as the previous
    value.

    >>> filt = rt.FA([True, True, False, True, True, False, False, True, True, True])
    >>> rt.FastArray.ema_decay(data, times, 0.5, filter=filt)
    FastArray([1.        , 1.6065307 , 1.6065307 , 1.5910096 , 1.9649961 ,
            1.1918304 , 0.92819846, 1.0761912 , 2.04962   , 2.6366549 ],
            dtype=float32)

    A reset mask (boolean array) can be also be provided when using a filter.
    Each each position where the reset mask is True, the EMA value is reset to the
    corresponding element of `arr`.

    >>> reset_mask = rt.FA([False, True, False, False, False, False, False, False, False, True])
    >>> rt.FastArray.ema_decay(data, times, 0.5, filter=filt, reset=reset_mask)
    FastArray([1.        , 1.        , 1.        , 1.3678794 , 1.8296608 ,
            1.1097454 , 0.86427057, 1.0709436 , 2.044502  , 1.        ],
            dtype=float32)
    """
    if not isinstance(arr, np.ndarray):
        arr = np.asanyarray(arr)

    if decay_rate is None:
        raise ValueError("ema_decay function requires a kwarg 'decay_rate' floating point value as input")

    if time is None:
        raise ValueError('ema_decay function requires a time array.  Use the "time" kwarg')

    if not isinstance(time, np.ndarray):
        raise ValueError("ema_decay function requires a time numpy array.")

    # require: len(arr) == len(time)
    if arr.shape != time.shape:
        raise ValueError("ema_decay requires the `time` array to be the same shape as the `arr` array.")

    # Allocate the output array
    output = empty_like(arr, dtype=dtype)

    if filter is not None:
        if not isinstance(filter, np.ndarray):
            raise ValueError("ema_decay function requires a filter numpy array.")

        # require: len(arr) == len(filter)
        if arr.shape != filter.shape:
            raise ValueError(
                "ema_decay requires the `filter` array, when supplied, to be the same shape as the `arr` array."
            )

        if reset is not None:
            if not isinstance(reset, np.ndarray):
                raise ValueError("ema_decay function requires a reset numpy array.")

            # require: len(arr) == len(reset)
            if arr.shape != reset.shape:
                raise ValueError(
                    "ema_decay requires the `reset` array, when supplied, to be the same shape as the `arr` array."
                )

            nb_ema_decay_with_filter_and_reset(arr, output, time, decay_rate, filter, reset)
        else:
            nb_ema_decay_with_filter(arr, output, time, decay_rate, filter)
    else:
        # If a 'reset' was supplied by the user, raise a warning to notify the user the reset won't be applied since they didn't provide a filter.
        if reset is not None:
            raise UserWarning(
                "ema_decay will not apply the `reset` array to the calculation because a filter was not specified."
            )

        nb_ema_decay(arr, output, time, decay_rate)
    return output


# -------------------------------------------------------
# Keep at bottom of this file
FastArray.register_function("fill_forward", fill_forward)
FastArray.register_function("fill_backward", fill_backward)
FastArray.register_function("ema_decay", ema_decay)
FastArray.register_function("cummax", cummax)
FastArray.register_function("cummin", cummin)



import sys

import numpy as np
import pytest
import riptide_cpp as rc
from numpy.random import default_rng
from numpy.testing import assert_array_equal

import riptable as rt

# TODO: Implement tests for sorting, binning, ismember, hstack, etc.


class TestBasicMath:
    """Unit tests for the 'BasicMath' functions in riptide_cpp."""

    # TODO: Use pytest.parametrize() to make it easy to add additional cases for this test.
    #       Consider performing a combinatorial sweep over various inputs and functions (though maybe we'd run this under
    #       the long-running tests build instead then).
    def test_basicmath_two_inputs(self):
        arg0 = rt.FA([1, 2, 3, 4, 5]).tile(40)
        arg1 = np.array(1, dtype=np.int64)
        tupleargs = (arg0, arg1)

        final_num = 7
        fastfunction = rt.rt_enum.MATH_OPERATION.CMP_EQ

        # Try the operation.
        result = rc.BasicMathTwoInputs(tupleargs, fastfunction, final_num)

        assert_array_equal(result, rt.FA([True, False, False, False, False]).tile(40))


class TestIndexing:
    """Unit tests for various indexing functions in riptide_cpp."""

    def test_boolindex(self) -> None:
        # rc.BooleanIndex(...)
        pytest.skip("Test not yet implemented.")

    @pytest.mark.skipif(
        sys.platform != "win32",
        reason="This test fails on Linux, perhaps due to int/long/longlong mismatch -- need to investigate.",
    )
    def test_mbget_int32_int64(self) -> None:
        arg0 = rt.FA([1, 2, 3, 4, 5, 6], dtype=np.int32)
        arg1 = rt.FA(
            [
                -9223372036854775808,
                -9223372036854775808,
                -9223372036854775808,
                0,
                3,
                -9223372036854775808,
                1,
                4,
                -9223372036854775808,
                2,
                5,
                -9223372036854775808,
            ],
            dtype=np.int64,
        )

        # Try the operation
        result = rc.MBGet(arg0, arg1)
        arr_inv = arg0.inv
        expected = rt.FA([arr_inv, arr_inv, arr_inv, 1, 4, arr_inv, 2, 5, arr_inv, 3, 6, arr_inv])
        assert_array_equal(result, expected)

    @pytest.mark.skipif(
        sys.platform != "win32",
        reason="This test fails on Linux, perhaps due to int/long/longlong mismatch -- need to investigate.",
    )
    def test_mbget_bytes_int64(self) -> None:
        arg0 = rt.FA(["x", "y", "z", "q", "w", "t"])
        arg1 = rt.FA(
            [rt.int64.inv, rt.int64.inv, rt.int64.inv, 0, 3, rt.int64.inv, 1, 4, rt.int64.inv, 2, 5, rt.int64.inv],
            dtype=np.int64,
        )

        # Try the operation
        result = rc.MBGet(arg0, arg1)
        arr_inv = arg0.inv
        expected = rt.FA([arr_inv, arr_inv, arr_inv, b"x", b"q", arr_inv, b"y", b"w", arr_inv, b"z", b"t", arr_inv])
        assert_array_equal(result, expected)


class TestConversions:
    """Unit tests for conversion functions in riptide_cpp."""

    def test_convert_safe(self) -> None:
        # rc.ConvertSafe(...)
        pytest.skip("Test not yet implemented.")

    def test_convert_unsafe(self) -> None:
        # rc.ConvertUnsafe(...)
        pytest.skip("Test not yet implemented.")


class TestReductions:
    """Unit tests for reduction functions in riptide_cpp."""

    def test_reduce(self) -> None:
        # rc.Reduce(...)
        pytest.skip("Test not yet implemented.")


class TestLedgerFunctions:
    """Unit tests for ledger functions."""

    def test_foo(self) -> None:
        # rc.LedgerFunction(...)
        pytest.skip("Test not yet implemented.")


class TestCRC:
    """Unit tests for the CRC32/CRC64 calculation on arrays."""

    def test_raises_on_bad_arg_type(self) -> None:
        """Test that rc.CalculateCRC raises an error when called with the wrong argument type."""
        bytes = b"abcdefghi"
        with pytest.raises(TypeError):
            rc.CalculateCRC(bytes)

    def test_disallow_noncontig(self) -> None:
        """Test that rc.CalculateCRC raises an error when calling with a non-contiguous array."""
        arr = np.arange(100)
        with pytest.raises(ValueError):
            rc.CalculateCRC(arr[::2])

    def test_basic(self) -> None:
        # Array size (in bytes, since that's what CRC is going to look at).
        # This is a reasonably large, PRIME value so we make sure all code paths
        # are exercised (e.g. don't want to fall into some special case that
        # only handles arrays smaller than 1000 bytes or something).
        array_size_bytes = 82727

        # Create a "random" array (but seed the RNG so this test is deterministic).
        rng = default_rng(123456)
        bytes = np.frombuffer(rng.bytes(array_size_bytes), dtype=np.int8)

        # Calculate the CRC32C of several slices of the array.
        # We'll use these values to check if we're seeing hash-like cascading behavior;
        # this also allows us to avoid test breakage if the RNG algorithm changes
        # or otherwise has differences between numpy versions or systems.
        results = np.empty(
            shape=500, dtype=np.uint64
        )  # NOTE: dtype needs to be set based on the expected output type of the CRC
        for i in range(len(results)):
            # Calculate the CRC32C or CRC64 of the array.
            crc_result = rc.CalculateCRC(bytes[i:])
            results[i] = crc_result

        # What's the average hash value? It should be roughly half the range of a uint32.
        avg_hash = results.mean()
        uniform_mean = (
            float(np.iinfo(results.dtype).max) / 2.0
        )  # Mean of a discrete uniform distribution over [0, results.dtype.max]
        avg_hash_accuracy = avg_hash / uniform_mean

        # How far off is the hash avg from the midpoint of uint32?
        # It should only be by a few % (which should be lower / closer to the midpoint as the sample count increases).
        hash_avg_error = abs(1.0 - avg_hash_accuracy)
        assert (
            hash_avg_error < 0.05
        ), f"The mean of the calculated hash values differs from the expected mean {np.dtype(results.dtype).name}.max/2.0 by {hash_avg_error * 100}%."

        # Check the variance too to see if it's reasonably close to
        # what we'd expect from a uniform distribution.
        hash_var = results.astype(np.float64).var()
        # This is population variance (not sample var), but it's good enough for this test.
        uniform_variance = (float(np.iinfo(np.uint64).max) ** 2 - 1) / 12.0
        hash_var_accuracy = hash_var / uniform_variance

        hash_var_error = abs(1.0 - hash_var_accuracy)
        assert (
            hash_var_error < 0.10
        ), f"The variance of the calculated hash values differs from the expected variance by {hash_var_error * 100}%."



__all__ = ["PDataset"]

import os
import warnings
from typing import List, Union

import numpy as np

from .rt_dataset import Dataset
from .rt_enum import DisplayJustification, TypeRegister
from .rt_fastarray import FastArray
from .rt_groupby import GroupBy
from .rt_itemcontainer import ItemContainer
from .rt_numpy import cumsum, empty, max, searchsorted, unique
from .rt_sds import load_sds


class PDataset(Dataset):
    """
    The PDataset class inherits from Dataset.  It holds multiple datasets (preivously stacked together) in contiguous slices.
    Each partition has a name and a contiguous slice that can be used to extract it from the larger Dataset.
    Extracting a partition is zero-copy.  Partitions can be extracted using partition(), or bracket [] indexing.

    A PDataset is often returned when:
        Multiple Datasets are hstacked, i.e. hstack([ds1, ds2, ds3])
        Calling load_sds with stack=True, i.e. load_sds([file1, file2, file3], stack=True)

    Properties: prows, pdict, pnames, pcount, pgb, pgbu, pgroupby, pslices, piter, pcutoffs
    Methods: partition(), pslice(), showpartitions()

    pds['20190204']  or  pds[20190204]  will return a dataset for the given partition name

    **Construction**

    inputval : -list of files to load and stack
               -list of datasets to stack
               -regular dataset inputval (will only have one partition)

    PDataset([path1, path2, path3], (pnames))
    -call load_sds(stack=True)
    -paths become filenames
    -if pnames specified, use those, otherwise look for dates
    -if no dates, auto generate pnames

    PDataset([ds1, ds2, ds3], (filenames, pnames))
    PDataset(ds, (filenames, pnames))
    -call Dataset.hstack()
    -if pnames specified, use those
    -if filenames, look for dates
    -if no dates, auto generate pnames

    PDataset(arraydict, cutoffs, (filenames, pnames))
    -constructor from load_sds()
    -if pnames specified, use those
    -if filenames, look for dates
    -if no dates, auto generate pnames

    """

    # ------------------------------------------------------------
    def __init__(
        self,
        inputval: Union[list, dict, "Dataset", "ItemContainer"] = None,
        cutoffs=None,
        filenames: List[str] = None,
        pnames=None,
        showpartitions=True,
        **kwargs,
    ):
        if inputval is None:
            inputval = dict()
        if filenames is None:
            filenames = list()
        if type(inputval) == TypeRegister.Dataset:
            inputval = [inputval]

        # stack datasets or load from list of files
        if isinstance(inputval, list):
            inputval, cutoffs, filenames, pnames = self._init_from_list(inputval, filenames, pnames)

        self._pre_init()

        # fast track for itemcontainer
        if isinstance(inputval, ItemContainer):
            self._init_from_itemcontainer(inputval)
        # load items from object that can be turned into dictionary
        else:
            inputval = self._init_columns_as_dict(inputval)
            self._init_from_dict(inputval)

        self._post_init(
            cutoffs=cutoffs,
            filenames=filenames,
            pnames=pnames,
            showpartitions=showpartitions,
        )

    # ------------------------------------------------------------
    def _pre_init(self):
        """
        Keep this in for chaining pre-inits in parent classes.
        """
        super()._pre_init()

    # ------------------------------------------------------------
    def _post_init(self, cutoffs, filenames, pnames, showpartitions):
        """
        Final initializer for variables specific to PDataset.
        Also initializes variables from parent class.
        """
        super()._post_init()

        self._showpartitions = showpartitions

        # cutoffs will be the same for dataset columns
        if cutoffs is not None:
            self._pcutoffs = list(cutoffs.values())[0]
        else:
            # assume one row, init from dataset
            self._pcutoffs = FastArray([self._nrows])

        # number of rows in each partition
        self._prows = self._pcutoffs.copy()

        if len(self._prows) > 1:
            # calculate row length
            self._prows[1:] -= self._pcutoffs[:-1]

        # look for dates in filenames or autogenerate names
        if pnames is None:
            pnames, filenames = self._init_pnames_filenames(len(self._prows), pnames, filenames)
            self._pfilenames = filenames
            self._pnames = {p: i for i, p in enumerate(pnames)}
        # use provided pnames
        else:
            self._pfilenames = filenames
            if isinstance(pnames, list):
                pnames = {p: i for i, p in enumerate(pnames)}
            self._pnames = pnames

        self._pcat = None

    # ------------------------------------------------------------
    @classmethod
    def _filenames_to_pnames(cls, filenames):
        """
        At least two filenames must be present to compare
        Algo will reverse the string on the assumption that pathnames can vary in the front of the string
        It also assumes that the filenames end similarly, such as ".SDS"
        It will search for the difference and look for digits, then try to extract the digits
        """
        # reverse all the filenames
        if len(filenames) > 0:
            rfilenames = [f[::-1] for f in filenames]

            str_arr = TypeRegister.FastArray(rfilenames)
            str_numba = str_arr.numbastring

            if len(filenames) > 1:
                match_mask = str_numba[0] != str_numba[1]

                str_len = len(match_mask)

                for i in range(len(filenames) - 2):
                    # inplace OR loop so that the TRUE propagates
                    match_mask += str_numba[0] != str_numba[i + 2]

                for i in range(str_len):
                    if match_mask[i]:
                        break

                start = i

                for i in range(start + 1, str_len):
                    if not match_mask[i]:
                        break
                end = i

                # expand start if possible
                while start > 0:
                    char = str_numba[0][start - 1]
                    # as long as a numeric digit, keep expanding
                    if char >= 48 and char <= 58:
                        start = start - 1
                    else:
                        break

                # expand end if possible
                while end < str_len:
                    char = str_numba[0][end]
                    if char >= 48 and char <= 58:
                        end = end + 1
                    else:
                        break

                # check to see if we captured a number
                firstchar = str_numba[0][start]
                lastchar = str_numba[0][end - 1]
                if start < end and firstchar >= 48 and firstchar <= 58 and lastchar >= 48 and lastchar <= 58:
                    pnames = []
                    viewtype = "S" + str(end - start)
                    for i in range(len(filenames)):
                        newstring = str_numba[i][start:end].view(viewtype)
                        newstring = newstring[0].astype("U")
                        # append the reverse
                        pnames.append(newstring[::-1])

                    u = unique(pnames)
                    if len(u) == len(filenames):
                        return pnames
                    # removed, prints during every column index/copy
                    # print(f"Failed to find unique numbers in filenames {pnames}")
            else:
                # only one file
                filename = str(rfilenames[0])
                start = -1
                stop = -1
                # search for first number
                for i in range(len(filename)):
                    if filename[i].isdigit():
                        if start == -1:
                            start = i
                    elif start != -1:
                        stop = i
                        break

                if start != -1:
                    if stop == -1:
                        stop = start + 1

                    # extract just the number
                    filename = filename[start:stop]
                    return [filename[::-1]]

        # failed to find unique strings in filenames
        # default to  p0, p1, p2
        pnames = cls._auto_pnames(len(filenames))

        return pnames

    # ------------------------------------------------------------
    @classmethod
    def _init_from_list(cls, dlist, filenames, pnames):
        """
        Construct a PDataset from multiple datasets, or by loading multiple files.
        """
        # make sure only one type
        listtype = {type(i) for i in dlist}
        if len(listtype) == 1:
            listtype = list(listtype)[0]
        else:
            raise TypeError(f"Found multiple types in constructor list {listtype}")

        # hstack datasets
        if listtype == Dataset:
            start = 0
            cutoffs = cumsum([ds.shape[0] for ds in dlist])
            cutoffs = {"cutoffs": cutoffs}
            ds = TypeRegister.Dataset.concat_rows(dlist)
            # extract itemcontainer
            ds = ds._all_items
            pnames, filenames = cls._init_pnames_filenames(len(dlist), pnames, filenames)

        # perform a .sds load from multiple files
        elif issubclass(listtype, (str, bytes, os.PathLike)):
            ds = load_sds(dlist, stack=True)
            cutoffs = {"cutoffs": ds._pcutoffs}
            filenames = ds._pfilenames
            if pnames is None:
                pnames = ds._pnames  # dict
            # extract itemcontainer
            ds = ds._all_items

        else:
            raise TypeError(f"Cannot construct from list of type {listtype}")

        return ds, cutoffs, filenames, pnames

    # ------------------------------------------------------------
    @classmethod
    def _auto_pnames(cls, pcount):
        """
        Auto generate partition names if none provided and no date found in filenames.
        """
        return ["p" + str(i) for i in range(pcount)]

    # ------------------------------------------------------------
    def _autocomplete(self) -> str:
        return f"PDataset{self.shape}"

    # ------------------------------------------------------------
    @classmethod
    def _init_pnames_filenames(cls, pcount, pnames, filenames):
        """
        Initialize filenames, pnames based on what was provided to the constructor.

        If no pnames provided, try to derive a date from filenames
        If no date found, or no filenames provided, use default names [p0, p1, p2 ...]

        Parameters
        ----------
        pcount    : int
            number of partitions, in case names need to be auto generated
        pnames    : list of str, optional
            list of partition names or None
        filenames : sequence of str, optional
            list of file paths (possibly empty)
        """
        if pnames is None:
            if filenames is None or len(filenames) == 0:
                filenames = []
                pnames = cls._auto_pnames(pcount)
            else:
                pnames = cls._filenames_to_pnames(filenames)

        return pnames, filenames

    # ------------------------------------------------------------
    def _copy(self, deep=False, rows=None, cols=None, base_index=0, cls=None):
        """returns a PDataset if no row selection, otherwise Dataset"""

        if rows is None:
            newcols = self._as_itemcontainer(deep=deep, rows=rows, cols=cols, base_index=base_index)
            # create a new PDataset
            pds = type(self)(
                newcols,
                cutoffs={"cutoffs": self.pcutoffs},
                filenames=self._pfilenames,
                pnames=self._pnames,
                base_index=base_index,
            )
            pds = self._copy_attributes(pds, deep=deep)
        else:
            # row slicing will break partitions, return a regular Dataset
            cls = TypeRegister.Dataset
            pds = super()._copy(deep=deep, rows=rows, cols=cols, base_index=base_index, cls=cls)

        return pds

    # ------------------------------------------------------------
    def _ipython_key_completions_(self):
        # For tab autocomplete with __getitem__
        # NOTE: %config IPCompleter.greedy=True   might have to be set
        # autocompleter will sort the keys
        return self.keys() + self.pnames

    # ------------------------------------------------------------
    @property
    def pcutoffs(self):
        """
        Returns
        -------
        Cutoffs for partition. For slicing, maintain contiguous arrays.

        Examples
        --------
        >>> pds.pcutoffs
        FastArray([1447138, 3046565, 5344567], dtype=int64)
        """
        return self._pcutoffs

    # ------------------------------------------------------------
    @property
    def prows(self):
        """
        Returns
        -------
        An array with the number of rows in each partition.

        Examples
        --------
        Example below assumes 3 filenames date encoded with datasets

        >>> pds = load_sds([file1, file2, file3], stack=True)
        >>> pds.prows
        FastArray([1447138, 2599427, 1909895], dtype=int64)
        """
        return self._prows

    # ------------------------------------------------------------
    @property
    def pcount(self):
        """
        Returns
        -------
        Number of partitions

        Examples
        --------
        Example below assumes 3 filenames date encoded with datasets

        >>> pds = load_sds([file1, file2, file3], stack=True)
        >>> pds.pcount
        3
        """
        return len(self._prows)

    # ------------------------------------------------------------
    @property
    def pnames(self):
        """
        Returns
        -------
        A list with the names of the partitions

        Example
        -------
        Example below assumes 3 filenames date encoded with datasets

        >>> pds = load_sds([file1, file2, file3], stack=True)
        >>> pds.pnames
        ['20190205', '20190206', '20190207']
        """
        return [*self._pnames.keys()]

    def set_pnames(self, pnames):
        """
        Parameters
        ----------
        pnames: list of str

        Examples
        --------
        Example below assumes 3 filenames date encoded with datasets

        >>> pds = load_sds([file1, file2, file3], stack=True)
        >>> pds.pnames
        ['20190205', '20190206', '20190207']
        >>> pds.set_pnames(['Jane', 'John', 'Jill'])
        ['Jane', 'John', 'Jill']
        """
        if isinstance(pnames, list):
            if len(pnames) == len(self._pnames):
                newpnames = {}
                for i in range(len(pnames)):
                    newpnames[pnames[i]] = i

                if len(newpnames) == len(self._pnames):
                    self._pnames = newpnames
                else:
                    raise ValueError(f"The new pnames must be unique names: {pnames}")

            else:
                raise ValueError(
                    f"The length of the new pnames must match the length of the old pnames: {len(self._pnames)}"
                )
        else:
            raise ValueError(f"A list of string must be passed in")

        return [*self._pnames.keys()]

    # ------------------------------------------------------------
    @property
    def pdict(self):
        """
        Returns
        --------
        A dictionary with the partition names and the partition slices.

        Examples
        --------
        Example below assumes 3 filenames date encoded with datasets

        >>> pds = load_sds([file1, file2, file3], stack=True)
        >>> pds.pdict
        {'20190204': slice(0, 1447138, None),
         '20190205': slice(1447138, 3046565, None),
         '20190206': slice(3046565, 4509322, None)}
        """
        pdict = {name: self.pslice(i) for i, name in enumerate(self.pnames)}
        return pdict

    # ------------------------------------------------------------
    # -------------------------------------------------------
    def pgb(self, by, **kwargs):
        """Equivalent to :meth:`~rt.rt_dataset.Dataset.pgroupby`"""
        kwargs["sort"] = True
        return self.pgroupby(by, **kwargs)

    # -------------------------------------------------------
    def pgroupby(self, by, **kwargs):
        return GroupBy(self, by, cutoffs=self._pcutoffs, **kwargs)

    def igroupby(self):
        """
        Lazily generate a categorical binned by each partition.
        Data will be attached to categorical, so operations can be called without specifying data.
        This allows reduce functions to be applied per partion.

        Examples
        --------
        Example below assumes 3 filenames date encoded with datasets

        >>> pds = load_sds([file1,file2, file2], stack=True)
        >>> pds.pgroupby['AskSize'].sum()
        *Partition   TradeSize
        ----------   ---------
        20190204     1.561e+07
        20190205     1.950e+07
        20190206     1.532e+07

        See Also: Dataset.groupby, Dataset.gb, Dataset.gbu
        """
        reserved_name = "Partition"

        if reserved_name not in self:
            self[reserved_name] = self.pcat
            self.col_move_to_front(reserved_name)

        return self.gb(reserved_name)

    @property
    def pcat(self):
        """
        Lazy generates a categorical for row labels callback or pgroupby
        """
        if self._pcat is None:
            idx = empty((self.shape[0],), dtype=np.int32)
            for i in range(self.pcount):
                idx[self.pslice(i)] = i + 1

            label = self.pnames
            self._pcat = TypeRegister.Categorical(idx, label)
        return self._pcat

    # ------------------------------------------------------------
    def prow_labeler(self, rownumbers, style):
        """
        Display calls this routine back to replace row numbers.
        rownumbers   : fancy index of row numbers being displayed
        style : ColumnStyle object - default from DisplayTable, can be changed

        Returns: label header, label array, style
        """
        if self._showpartitions:
            style.align = DisplayJustification.Right

            # use the cutoffs to generate which partition index
            pindex = searchsorted(self._pcutoffs, rownumbers, side="right")
            plabels = TypeRegister.FastArray(self.pnames)[pindex]

            # find the maximum string width for the rownumber
            if len(rownumbers) > 0:
                maxnum = max(rownumbers)
            else:
                maxnum = 0
            width = len(str(maxnum))

            # right justify numbers
            rownumbers = rownumbers.astype("S")
            rownumbers = np.chararray.rjust(rownumbers, width)

            # column header
            header = "partition + #"
            rownumbers = plabels + " " + rownumbers

            # set the style width to override the string trim
            style.width = rownumbers.itemsize

            return header, rownumbers, style
        else:
            return "#", rownumbers, style

    # ------------------------------------------------------------
    @property
    def _row_numbers(self):
        # display will check for the existence of this method
        # return a callback to change the row numbers
        return self.prow_labeler

    # ------------------------------------------------------------
    def showpartitions(self, show=True):
        """toggle whether partitions are shown on the left"""
        if show:
            self._showpartitions = True
        else:
            self._showpartitions = False

    # ------------------------------------------------------------
    @property
    def piter(self):
        """
        Iterate over dictionary of arrays for each partition.
        Yields key (load source) -> value (dataset as dictionary)

        Examples
        --------
        Example below assumes 3 filenames date encoded with datasets

        >>> pds = load_sds([file1,file2, file2], stack=True)
        >>> for name, ds in pds.iter: print(name)
        20190204
        20190205
        20190206
        """

        label = self.pnames
        start = 0
        for i in range(self.pcount):
            yield label[i], self.partition(i)

    # -------------------------------------------------------
    @property
    def pslices(self):
        """
        Return the slice (start,end) associated with the partition number

        See Also
        --------
        pslices, pdict

        Examples
        --------
        Example below assumes 3 filenames date encoded with datasets

        >>> pds = load_sds([file1,file2, file2], stack=True)
        >>> pds.pslices
        [slice(0, 1447138, None),
         slice(1447138, 3046565, None),
         slice(3046565, 4509322, None)]
        """

        pslices = [self.pslice(i) for i in range(self.pcount)]
        return pslices

    # -------------------------------------------------------
    def pslice(self, index):
        """
        Return the slice (start,end) associated with the partition number

        See Also
        --------
        pslices, pdict

        Examples
        --------
        >>> pds.pslice(0)
        slice(0, 1447138, None)

        """
        if isinstance(index, (int, np.integer)):
            if index == 0:
                return slice(0, self.pcutoffs[index])
            else:
                return slice(self.pcutoffs[index - 1], self.pcutoffs[index])

        raise IndexError(f"Cannot slice a partition with type {type(index)!r}.  Use an integer instead.")

    # -------------------------------------------------------
    def partition(self, index):
        """
        Return the Dataset associated with the partition number

        Examples
        --------
        Example below assumes 3 filenames with datasets

        >>> pds = load_sds([file1, file2, file2], stack=True)
        >>> pds.partition(0)
        """

        if isinstance(index, (int, np.integer)):
            # this will route to the dataset
            return self._copy(rows=self.pslice(index))

        if isinstance(index, str):
            # this will loop back if the string is a partition name
            return self[index]

        raise IndexError(f"Cannot index a parition with type {type(index)!r}.  Use an integer instead.")

    # -------------------------------------------------------
    def __getitem__(self, index):
        """
        :param index: (rowspec, colspec) or colspec
        :return: the indexed row(s), cols(s), sub-dataset or single value
        :raise IndexError:
        :raise TypeError:
        :raise KeyError:
        """
        try:
            return super().__getitem__(index)
        except:
            # if it fails, maybe it was a partition selection
            if isinstance(index, (int, np.integer)):
                # convert int to string to lookup
                index = str(index)

            # the string was not a column name, now check for partition name
            if isinstance(index, str):
                if index in self._pnames:
                    # return the dataset for that partition
                    return self.partition(self._pnames[index])
                else:
                    raise KeyError(f"the key {index!r} was not found as column name or parition name")
            else:
                raise KeyError(f"could not index PDataset with {type(index)}")

    # --------------------------------------------------------------------------
    def save(
        self,
        path="",
        share=None,
        compress=True,
        overwrite=True,
        name=None,
        onefile: bool = False,
        bandsize=None,
        append=None,
        complevel=None,
    ):
        warnings.warn(f"To be implemented. PDataset will currently be saved / loaded as a Dataset.")
        super().save(
            path=path,
            share=share,
            compress=compress,
            overwrite=overwrite,
            name=name,
            onefile=onefile,
            bandsize=bandsize,
            append=append,
            complevel=complevel,
        )

    # --------------------------------------------------------------------------
    @classmethod
    def hstack(cls, pds_list):
        """
        Stacks columns from multiple datasets.
        see: Dataset.concat_rows
        """
        raise NotImplementedError("PDataset does not stack yet")

    # ------------------------------------------------------------
    @classmethod
    def pload(cls, path, start, end, include=None, threads=None, folders=None):
        """
        Returns a PDataset of stacked files from multiple days.
        Will load all files found within the date range provided.

        Parameters
        ----------
        path  : format string for filepath, {} in place of YYYYMMDD. {} may appear multiple times.
        start : integer or string start date in format YYYYMMDD
        end   : integer or string end date in format YYYYMMDD
        """
        # insert date string at each of these
        fmtcount = path.count("{}")

        # final loader will check if dates exist, kill warnings?
        pnames = TypeRegister.Date.range(str(start), str(end)).yyyymmdd.astype("U")

        files = [path.format(*[d] * fmtcount) for d in pnames]

        pds = load_sds(files, include=include, stack=True, threads=threads, folders=folders)

        return pds

    # ------------------------------------------------------------
    def psave(self):
        """
        Does not work yet.  Would save backout all the partitions.
        """
        raise NotImplementedError(f"not implemented yet")


TypeRegister.PDataset = PDataset



# -*- coding: utf-8 -*-
"""Benchmark functionality."""
from .bench_categorical import bench_multikey_cat_isin, bench_singlekey_cat_isin
from .bench_grouping import bench_groupbyhash, bench_groupbylex, bench_groupbypack
from .bench_merge import (
    bench_merge,
    bench_merge2,
    bench_merge2_with_settings,
    bench_merge_pandas,
    compare_merge,
)
from .bench_numpy import (
    bench_rt_add,
    bench_rt_all,
    bench_rt_all_with_nan,
    bench_rt_any,
    bench_rt_any_with_nan,
    bench_rt_arange,
    bench_rt_argmax,
    bench_rt_argmin,
    bench_rt_argsort,
    bench_rt_isnan,
    bench_rt_lexsort_ordered,
    bench_rt_lexsort_reversed,
    bench_rt_max,
    bench_rt_mean,
    bench_rt_median,
    bench_rt_min,
    bench_rt_minimum,
    bench_rt_nanmax,
    bench_rt_nanmean,
    bench_rt_nanmedian,
    bench_rt_nanmin,
    bench_rt_nanpercentile,
    bench_rt_nanstd,
    bench_rt_nansum,
    bench_rt_nanvar,
    bench_rt_percentile,
    bench_rt_sqrt,
    bench_rt_std,
    bench_rt_sum,
    bench_rt_true_divide,
    bench_rt_trunc,
    bench_rt_var,
    benchmark_binary,
    benchmark_reduce,
    benchmark_unary,
    compare_add,
    compare_all,
    compare_all_with_nan,
    compare_any,
    compare_any_with_nan,
    compare_argmax,
    compare_argmin,
    compare_isnan,
    compare_lexsort,
    compare_max,
    compare_mean,
    compare_median,
    compare_min,
    compare_minimum,
    compare_numpy_binary,
    compare_numpy_reduce,
    compare_numpy_unary,
    compare_percentile,
    compare_sqrt,
    compare_std,
    compare_sum,
    compare_true_divide,
    compare_trunc,
    compare_var,
)
from .bench_primops import (
    bench_astype,
    bench_astype_numba,
    bench_bool_index,
    bench_compare_ops,
    bench_compare_ops_numpy,
    bench_mbget,
    bench_mbget_numba,
    compare_astype,
    compare_compare_ops,
    compare_mbget,
)
from .benchmark import benchmark

__all__ = [
    "benchmark",
    "benchmark_binary",
    "benchmark_reduce",
    "benchmark_unary",
    "bench_groupbyhash",
    "bench_groupbylex",
    "bench_groupbypack",
    "bench_mbget",
    "bench_mbget_numba",
    "bench_astype",
    "bench_astype_numba",
    "bench_compare_ops",
    "bench_compare_ops_numpy",
    "compare_mbget",
    "compare_astype",
    "compare_compare_ops",
    "bench_bool_index",
    "bench_merge",
    "bench_merge2",
    "bench_merge2_with_settings",
    "bench_merge_pandas",
    "bench_singlekey_cat_isin",
    "bench_multikey_cat_isin",
    "bench_rt_all",
    "bench_rt_all_with_nan",
    "bench_rt_any",
    "bench_rt_any_with_nan",
    "bench_rt_arange",
    "bench_rt_argmax",
    "bench_rt_argmin",
    "bench_rt_argsort",
    "bench_rt_lexsort_ordered",
    "bench_rt_lexsort_reversed",
    "bench_rt_sum",
    "bench_rt_max",
    "bench_rt_min",
    "bench_rt_std",
    "bench_rt_add",
    "bench_rt_minimum",
    "bench_rt_nanmax",
    "bench_rt_nanmin",
    "bench_rt_nansum",
    "bench_rt_true_divide",
    "bench_rt_isnan",
    "bench_rt_trunc",
    "bench_rt_sqrt",
    "bench_rt_mean",
    "bench_rt_nanmean",
    "bench_rt_median",
    "bench_rt_nanmedian",
    "bench_rt_percentile",
    "bench_rt_nanpercentile",
    "bench_rt_var",
    "bench_rt_nanvar",
    "bench_rt_nanstd",
    "compare_all",
    "compare_all_with_nan",
    "compare_any",
    "compare_any_with_nan",
    "compare_argmax",
    "compare_argmin",
    "compare_lexsort",
    "compare_merge",
    "compare_sum",
    "compare_max",
    "compare_min",
    "compare_std",
    "compare_add",
    "compare_mean",
    "compare_median",
    "compare_minimum",
    "compare_true_divide",
    "compare_isnan",
    "compare_trunc",
    "compare_sqrt",
    "compare_var",
    "compare_numpy_reduce",
    "compare_numpy_binary",
    "compare_numpy_unary",
]
